import numpy as np
import FINITO_FEM_TOOLBOX.FINITO_COMMON_LIBRARY as FINITO_CL
import FINITO_FEM_TOOLBOX.FINITO_MEF1D_LIBRARY as FINITO_MEF1D
import FINITO_FEM_TOOLBOX.FINITO_MEF2D_LIBRARY as FINITO_MEF2D

def MEF1D(**kwargs):
    """ 
    This function performs structural analysis of frame elements with 2 nodes (1 at each end).
  
    Input:
    All inputs kwargs arguments type.

    FILENAME    | Structural dataset                                                 | .txt extension
    DICTIONARY  | Structural dataset                                                 | Py dictionary
                |   Dictionary and file tags                                         |
                |    TYPE_ELEMENT  = Type element in Finito algorithm                | Integer 
                |     0 - Frame bar element                                          |
                |    TYPE_SOLUTION = Boundary conditions in system of equations      | Integer
                |     0 - Condense procedure                                         |
                |     1 - Zero and One algorithm                                     |
                |   N_NODES        = Number of nodes                                 | Integer
                |   N_MATERIALS    = Number of materials                             | Integer
                |   N_SECTIONS     = Number of sections                              | Integer
                |   N_ELEMENTS     = Number of frame elements                        | Integer
                |   N_DOFPRESCRIPTIONS = Number of DOF displacement control          | Integer
                |   N_DOFLOADED        = Number of DOF forces                        | Integer
                |   N_DOFSPRINGS       = Number of DOF spring elements               | Integer
                |   COORDINATES        = Coordinates properties                      | Py Numpy array
                |                          Node, x, y                                |
                |   ELEMENTS       = Elements properties                             | Py Numpy array
                |                    Node 0 ... Node (N_NODES - 1), Material ID,     | 
                |                    Geometry ID, Hinge ID node 0, Hinge ID node 1   |
                |   MATERIALS      = Materials properties                            | Py Numpy array
                |                    Young, Poisson, Density, Thermal coefficient    |
                |   SECTIONS       = Sections properties                             | Py Numpy array
                |                    Area, Inertia 1, Inertia Frame bar, X GC, Y GC  |
                |   PRESCRIPTIONS  = Prescribed DOF displacement properties          | Py Numpy array              
                |                    Node, Direction (X = 0, Y = 1, Z = 2), Value    | 
                |   NODAL_LOAD     = Nodal DOF force properties                      | Py Numpy array              
                |                    Node, Direction (X = 0, Y = 1, Z = 2), Value    |
                |   SPRINGS        = Nodal DOF spring properties                     | Py Numpy array              
                                     Node, Direction (X = 0, Y = 1, Z = 2), Value    |

    Output:
    RESULTS     | Structural analysis results by element                             | Py dictionary
    """
    # Read input file 
    FILENAME = kwargs.get('FILENAME')
    DICTIONARY = kwargs.get('DICTIONARY')
    if FILENAME:
        TYPE_SOLUTION, TYPE_ELEMENT, N_NODES, N_MATERIALS, N_SECTIONS, N_ELEMENTS, N_DOFPRESCRIPTIONS, N_DOFLOADED, N_DOFSPRINGS, COORDINATES, ELEMENTS, MATERIALS, SECTIONS, PRESCRIPTIONS, NODAL_LOAD, SPRINGS = FINITO_MEF1D.GET_VALUE_FROM_TXT_MEF1D_FINITO(FILENAME)
    else:
        TYPE_SOLUTION, TYPE_ELEMENT, N_NODES, N_MATERIALS, N_SECTIONS, N_ELEMENTS, N_DOFPRESCRIPTIONS, N_DOFLOADED, N_DOFSPRINGS, COORDINATES, ELEMENTS, MATERIALS, SECTIONS, PRESCRIPTIONS, NODAL_LOAD, SPRINGS = FINITO_MEF1D.GET_VALUE_FROM_DICT_MEF1D_FINITO(DICTIONARY)
    # Creating the algorithm's internal parameters
    N_DOFSNODE, N_NODESELEMENT, DOFS_LOCAL, AUX_1, AUX_2, N_DOFSELEMENT = FINITO_CL.INDEX_ASSEMBLY(TYPE_ELEMENT)
    # Global DOF assembly
    DOF_GLOBALNODAL = FINITO_CL.DOF_GLOBAL_ASSEMBLY(TYPE_ELEMENT, N_DOFSNODE, N_NODES)
    # DOF's total, prescriptions and free
    DOF_GLOBAL, N_DOFSGLOBAL = FINITO_CL.TOTAL_DEGREE_FREEDOM(N_DOFSNODE, N_NODES)
    DOF_PRESCRIPTIONS, DOF_PRESCRIPTIONSVALUE, N_DOFSPRESCRIPTIONS = FINITO_CL.PRESCRIPTIONS_DEGREE_FREEDOM(PRESCRIPTIONS, DOF_GLOBALNODAL)
    DOF_FREE, N_DOFSFREE = FINITO_CL.FREE_DEGREE_FREEDOM(DOF_PRESCRIPTIONS, DOF_GLOBAL)
    # Nodal load contribuition
    DOF_NODALFORCE = FINITO_CL.CONTRIBUTION_NODAL_EXTERNAL_LOAD(NODAL_LOAD, N_DOFSGLOBAL, DOF_GLOBALNODAL)
    F_G = DOF_NODALFORCE
    # Structure stiffness matrix
    K_G = np.zeros((N_DOFSGLOBAL, N_DOFSGLOBAL))
    F_INT = np.zeros((N_DOFSGLOBAL, 1))
    # Hinged elements
    HINGES = FINITO_MEF1D.HINGED_PROPERTIES(ELEMENTS)
    for I_ELEMENT in range(N_ELEMENTS):
        # Material and section properties in j element
        MATERIAL_IELEMENT = FINITO_MEF1D.MATERIALS_PROPERTIES_0(ELEMENTS, MATERIALS, I_ELEMENT, AUX_1)
        SECTION_IELEMENT = FINITO_MEF1D.GEOMETRIC_PROPERTIES_0(COORDINATES, ELEMENTS, SECTIONS, I_ELEMENT, AUX_2)
        # i element stiffness matrix local axis
        K_IELEMENT = FINITO_MEF1D.ELEMENT_STIFFNESS_0(TYPE_ELEMENT, SECTION_IELEMENT, MATERIAL_IELEMENT, HINGES[I_ELEMENT, :])
        # Rotation matrix
        R_IELEMENT = FINITO_MEF1D.ELEMENT_ROTATION(TYPE_ELEMENT, SECTION_IELEMENT)
        # i element stiffness matrix global axis
        K_IELEMENTGLOBAL = np.dot(np.dot(np.transpose(R_IELEMENT), K_IELEMENT), R_IELEMENT)
        # Global DOF i element
        DOF_GLOBALIELEMENT = FINITO_CL.GLOBAL_DOF_ELEMENT(N_NODESELEMENT, N_DOFSNODE, DOF_GLOBALNODAL, ELEMENTS, I_ELEMENT)
        # Global stiffness
        K_GCONTRIBUITION = FINITO_CL.GLOBAL_STIFFNESS(N_DOFSGLOBAL, DOF_GLOBALIELEMENT, K_IELEMENTGLOBAL)
        K_G = K_G + K_GCONTRIBUITION
    # Spring contribution
    if N_DOFSPRINGS > 0:
        SPRING_INDEX, SPRING_VALUES = FINITO_CL.SPRING_CONTRIBUTION(N_DOFSNODE, SPRINGS, N_DOFSPRINGS)
        for K_COUNT in range(N_DOFSPRINGS):
            SPRING_VALUE = SPRING_VALUES[K_COUNT]
            INDEX_DOF = SPRING_INDEX[K_COUNT]
            K_G[INDEX_DOF, INDEX_DOF] = K_G[INDEX_DOF, INDEX_DOF] + SPRING_VALUE
    # Displacement solution 0: Condense procedure
    if TYPE_SOLUTION == 0:
        # Condense displacements, forces and Stiffness matrix
        U_PP = FINITO_CL.CONDENSE_PRESCRIBED_GLOBAL_DISPLACEMENT(DOF_PRESCRIPTIONSVALUE, N_DOFSPRESCRIPTIONS)
        F_FF = FINITO_CL.CONDENSE_FREE_GLOBAL_FORCES(F_G, DOF_FREE, N_DOFSFREE)
        K_FF = FINITO_CL.CONDENSE_GLOBAL_FREE_STIFFNESS(K_G, DOF_FREE, N_DOFSFREE)
        K_PF = FINITO_CL.CONDENSE_PRESCRIBED_FREE_GLOBAL_STIFFNESS(K_G, DOF_FREE, N_DOFSFREE, DOF_PRESCRIPTIONS, N_DOFSPRESCRIPTIONS)
        # Displacement solution
        K_FFINVERSE = np.linalg.pinv(K_FF, rcond = 1e-15)
        U_FF = np.dot(K_FFINVERSE, F_FF - np.dot(np.transpose(K_PF), U_PP))
        U_G = FINITO_CL.ASSEMBLY_TOTAL_DISPLACEMENT(U_FF, U_PP, N_DOFSGLOBAL, DOF_PRESCRIPTIONS, DOF_FREE)
    # Displacement solution 0: Zero and One procedure
    elif TYPE_SOLUTION == 1:
        K_G, F_G = FINITO_CL.ZERO_AND_ONE_METHOD(K_G, F_G, DOF_PRESCRIPTIONS, DOF_PRESCRIPTIONSVALUE)
        K_FFINVERSE = np.linalg.pinv(K_G, rcond = 1e-15)
        U_G = np.dot(K_FFINVERSE, F_G)
    # Internal loads
    # Frame division 
    DIV = 11 
    # Start empty dictionary
    RESULTS = [{'X': np.empty(DIV), 'UX': np.empty(DIV), 'UY': np.empty(DIV), 'UZ': np.empty(DIV), 'N': np.empty(DIV), 'V': np.empty(DIV), 'M': np.empty(DIV), 'ID_ELEMENT': J_COUNT} for J_COUNT in range(N_ELEMENTS)]
    for J_ELEMENT in range(N_ELEMENTS):
        # Material and section properties in j element
        MATERIAL_JELEMENT = FINITO_MEF1D.MATERIALS_PROPERTIES_0(ELEMENTS, MATERIALS, J_ELEMENT, AUX_1)
        SECTION_JELEMENT = FINITO_MEF1D.GEOMETRIC_PROPERTIES_0(COORDINATES, ELEMENTS, SECTIONS, J_ELEMENT, AUX_2)
        # j element stiffness matrix local axis 
        K_JELEMENT = FINITO_MEF1D.ELEMENT_STIFFNESS_0(TYPE_ELEMENT, SECTION_JELEMENT, MATERIAL_JELEMENT, HINGES[J_ELEMENT, :])
        # Rotation matrix
        R_JELEMENT = FINITO_MEF1D.ELEMENT_ROTATION(TYPE_ELEMENT, SECTION_JELEMENT)
        # Global DOF j element
        DOF_GLOBALJELEMENT = FINITO_CL.GLOBAL_DOF_ELEMENT(N_NODESELEMENT, N_DOFSNODE, DOF_GLOBALNODAL, ELEMENTS, J_ELEMENT)
        # Element global displacements
        U_GJELEMENT = FINITO_CL.CONDENSE_GLOBAL_ELEMENT_DISPLACEMENTS(U_G, N_DOFSELEMENT, DOF_GLOBALJELEMENT)
        # Element local displacements
        U_JELEMENT = np.dot(R_JELEMENT, U_GJELEMENT)
        # Internal force in j element (Node 0 and node 1 [x, y, tetha])
        F_ELINT = np.dot(K_JELEMENT, U_JELEMENT)
        # Internal force in j element (by division)
        for I_COUNT in range(DIV):
            # Local axis value
            X = SECTION_JELEMENT[0] * (I_COUNT) / (DIV - 1)
            if I_COUNT == 0:
                U_X = U_GJELEMENT[0, 0]
                U_Y = U_GJELEMENT[1, 0]
                U_Z = U_GJELEMENT[2, 0]
            elif I_COUNT == DIV - 1:
                U_X = U_GJELEMENT[3, 0]
                U_Y = U_GJELEMENT[4, 0]
                U_Z = U_GJELEMENT[5, 0]
            elif (I_COUNT != 0 and I_COUNT != DIV - 1):
                U_X = -1989
                U_Y = -1989
                U_Z = -1989
            # Internal loads: Axial, Shear and Bending Moment 
            N = -F_ELINT[0]
            V = F_ELINT[1]
            M = -F_ELINT[2] + F_ELINT[1] * X         
            # Save results in dictioonary
            RESULTS[J_ELEMENT]['X'][I_COUNT] = X
            RESULTS[J_ELEMENT]['UX'][I_COUNT] = U_X
            RESULTS[J_ELEMENT]['UY'][I_COUNT] = U_Y
            RESULTS[J_ELEMENT]['UZ'][I_COUNT] = U_Z
            RESULTS[J_ELEMENT]['N'][I_COUNT] = N
            RESULTS[J_ELEMENT]['V'][I_COUNT] = V
            RESULTS[J_ELEMENT]['M'][I_COUNT] = M
    return RESULTS

def MEF2D(**kwargs):
    """ 
    This function performs structural analysis via finite elements conside-
    ring flat surface elements (CST)

    Input:
    All inputs kwargs arguments type.

    FILENAME    | Structural dataset                      | txt extension
    DICTIONARY  | Structural dataset                      | Py dictionary

    Output:
    RESULTS     | Structural analysis results by node     | Py dictionary
    """
    # Read input file
    FILENAME = kwargs.get('FILENAME')
    DICTIONARY = kwargs.get('DICTIONARY')
    if FILENAME:
        [N_NODES, N_MATERIALS, N_THICKNESS, N_ELEMENTST3, N_ELEMENTST6, N_ELEMENTST10, N_FORCES, N_PRESSURES, N_DISPLACEMENTS, TYPE_PLANE, TYPE_ELEMENT, TYPE_SOLUTION, TYPE_INTEGRATION, COORDINATES, MATERIALS, THICKNESS, ELEMENTS, NODAL_LOAD, PRESCRIPTIONS] = FINITO_CL.GET_VALUE_FROM_TXT_MEF2D_FINITO(FILENAME)
    else:
        pass
    # Creating the algorithm's internal parameters
    N_DOFSNODE, N_NODESELEMENT, DOFS_LOCAL, AUX_1, AUX_2, N_DOFSELEMENT = FINITO_CL.INDEX_ASSEMBLY(TYPE_ELEMENT)
    # Global DOF assembly
    DOF_GLOBALNODAL = FINITO_CL.DOF_GLOBAL_ASSEMBLY(TYPE_ELEMENT, N_DOFSNODE, N_NODES)
    # DOF's total, prescriptions and free
    DOF_GLOBAL, N_DOFSGLOBAL = FINITO_CL.TOTAL_DEGREE_FREEDOM(N_DOFSNODE, N_NODES)
    DOF_PRESCRIPTIONS, DOF_PRESCRIPTIONSVALUE, N_DOFSPRESCRIPTIONS = FINITO_CL.PRESCRIPTIONS_DEGREE_FREEDOM(PRESCRIPTIONS, DOF_GLOBALNODAL)
    DOF_FREE, N_DOFSFREE = FINITO_CL.FREE_DEGREE_FREEDOM(DOF_PRESCRIPTIONS, DOF_GLOBAL)
    # Nodal load contribuition
    DOF_NODALFORCE = FINITO_CL.CONTRIBUTION_NODAL_EXTERNAL_LOAD(NODAL_LOAD, N_DOFSGLOBAL, DOF_GLOBALNODAL)
    F_G = DOF_NODALFORCE
    # Structure stiffness matrix
    K_G = np.zeros((N_DOFSGLOBAL, N_DOFSGLOBAL))
    # Numerical integration properties
    NUM_INT = FINITO_CL.NUMERICAL_INTEGRATION(TYPE_INTEGRATION)
    for I_ELEMENT in range(N_ELEMENTST3):
        # Material and section properties in i element
        C_IELEMENT = FINITO_CL.CONSTITUTIVE_C(TYPE_PLANE, MATERIALS, ELEMENTS, I_ELEMENT)
        SECTION_IELEMENT = FINITO_CL.GEOMETRIC_PROPERTIES_1(COORDINATES, ELEMENTS, I_ELEMENT, THICKNESS, AUX_2)
        # i stiffness matrix
        K_IELEMENTGLOBAL = FINITO_CL.ELEMENT_STIFFNESS_1(NUM_INT, N_DOFSELEMENT, TYPE_ELEMENT, N_NODESELEMENT, C_IELEMENT, SECTION_IELEMENT)
        # Global DOF in i element
        DOF_GLOBALIELEMENT = FINITO_CL.GLOBAL_DOF_ELEMENT(N_NODESELEMENT, N_DOFSNODE, DOF_GLOBALNODAL, ELEMENTS, I_ELEMENT)
        K_GCONTRIBUITION = FINITO_CL.GLOBAL_STIFFNESS(N_DOFSGLOBAL, DOF_GLOBALIELEMENT, K_IELEMENTGLOBAL)
        K_G = K_G + K_GCONTRIBUITION
    if TYPE_SOLUTION == 1:
        K_G, F_G = FINITO_CL.ZERO_AND_ONE_METHOD(K_G, F_G, DOF_PRESCRIPTIONS, DOF_PRESCRIPTIONSVALUE)
        U_G = np.linalg.solve(K_G, F_G)
    RESULTS = U_G
    return RESULTS                                                                                                                                                                                                                                                                                                                                                              
                                                                                                                                                                    
#   /$$$$$$  /$$$$$$$  /$$$$$$$$ /$$$$$$$$       /$$$$$$$$ /$$$$$$$$  /$$$$$$  /$$   /$$ /$$   /$$  /$$$$$$  /$$        /$$$$$$   /$$$$$$  /$$$$$$ /$$$$$$$$  /$$$$$$ 
#  /$$__  $$| $$__  $$| $$_____/| $$_____/      |__  $$__/| $$_____/ /$$__  $$| $$  | $$| $$$ | $$ /$$__  $$| $$       /$$__  $$ /$$__  $$|_  $$_/| $$_____/ /$$__  $$
# | $$  \__/| $$  \ $$| $$      | $$               | $$   | $$      | $$  \__/| $$  | $$| $$$$| $$| $$  \ $$| $$      | $$  \ $$| $$  \__/  | $$  | $$      | $$  \__/
# | $$ /$$$$| $$$$$$$/| $$$$$   | $$$$$            | $$   | $$$$$   | $$      | $$$$$$$$| $$ $$ $$| $$  | $$| $$      | $$  | $$| $$ /$$$$  | $$  | $$$$$   |  $$$$$$ 
# | $$|_  $$| $$____/ | $$__/   | $$__/            | $$   | $$__/   | $$      | $$__  $$| $$  $$$$| $$  | $$| $$      | $$  | $$| $$|_  $$  | $$  | $$__/    \____  $$
# | $$  \ $$| $$      | $$      | $$               | $$   | $$      | $$    $$| $$  | $$| $$\  $$$| $$  | $$| $$      | $$  | $$| $$  \ $$  | $$  | $$       /$$  \ $$
# |  $$$$$$/| $$      | $$$$$$$$| $$$$$$$$         | $$   | $$$$$$$$|  $$$$$$/| $$  | $$| $$ \  $$|  $$$$$$/| $$$$$$$$|  $$$$$$/|  $$$$$$/ /$$$$$$| $$$$$$$$|  $$$$$$/
#  \______/ |__/      |________/|________/         |__/   |________/ \______/ |__/  |__/|__/  \__/ \______/ |________/ \______/  \______/ |______/|________/ \______/ 


#  /$$$$$$$$ /$$$$$$ /$$   /$$ /$$$$$$ /$$$$$$$$  /$$$$$$        /$$$$$$$$  /$$$$$$   /$$$$$$  /$$       /$$$$$$$   /$$$$$$  /$$   /$$                                
# | $$_____/|_  $$_/| $$$ | $$|_  $$_/|__  $$__/ /$$__  $$      |__  $$__/ /$$__  $$ /$$__  $$| $$      | $$__  $$ /$$__  $$| $$  / $$                                
# | $$        | $$  | $$$$| $$  | $$     | $$   | $$  \ $$         | $$   | $$  \ $$| $$  \ $$| $$      | $$  \ $$| $$  \ $$|  $$/ $$/                                
# | $$$$$     | $$  | $$ $$ $$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$$$$$$ | $$  | $$ \  $$$$/                                 
# | $$__/     | $$  | $$  $$$$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$__  $$| $$  | $$  >$$  $$                                 
# | $$        | $$  | $$\  $$$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$  \ $$| $$  | $$ /$$/\  $$                                
# | $$       /$$$$$$| $$ \  $$ /$$$$$$   | $$   |  $$$$$$/         | $$   |  $$$$$$/|  $$$$$$/| $$$$$$$$| $$$$$$$/|  $$$$$$/| $$  \ $$                                
# |__/      |______/|__/  \__/|______/   |__/    \______/          |__/    \______/  \______/ |________/|_______/  \______/ |__/  |__/  

################################################################################
# UNIVERSIDADE FEDERAL DE CATALÃO (UFCAT)
# WANDERLEI MALAQUIAS PEREIRA JUNIOR,                  ENG. CIVIL / PROF (UFCAT)
# MARCOS NAPOLEÃO RABELO,                              ENG. CIVIL / PROF (UFCAT)
# DAVIDSON DE OLIVEIRA FRANÇA JUNIOR,                    ENG. CIVIL / PROF (UNA)
# GABRIEL BERNARDES CARVALHO,                                 ENG. CIVIL (UFCAT)
# JOSÉ VITOR CARVALHO SILVA,                                  ENG. CIVIL (UFCAT)
# MURILO CARNEIRO RODRIGUES,                                  ENG. CIVIL (UFCAT)
################################################################################

################################################################################
# DESCRIÇÃO ALGORITMO:
# BIBLIOTECA FINITO PARA FUNÇÕES DO ALGORITMO MEF1D DESENVOLVIDA PELO GRUPO DE 
# PESQUISA E ESTUDOS EM ENGENHARIA (GPEE)
################################################################################

################################################################################
# BIBLIOTECAS NATIVAS PYTHON
import numpy as np

################################################################################
# BIBLIOTECAS DESENVOLVEDORES GPEE

# LEITURA DO ARQUIVO TIPO .TXT
def GET_VALUE_FROM_TXT_MEF1D_FINITO(FILENAME):
    """
    This function reads data from .txt file.
    
    Input:
    FILENAME              | Structural dataset                                     | .txt extension

    Output: 
    TYPE_ELEMENT          | Type element in Finito algorithm                       | Integer 
                          |     0 - Frame bar element                              |
    TYPE_SOLUTION         | Solution of the system of equations                    | Integer
                          |    0 - Condense procedure                              |
                          |    1 - Zero and One algorithm                          |
    N_NODES               | Number of nodes                                        | Integer
    N_MATERIALS           | Number of materials                                    | Integer
    N_SECTIONS            | Number of sections                                     | Integer
    N_ELEMENTS            | Number of frame elements                               | Integer
    N_DOFPRESCRIPTIONS    | Number of DOF displacement control                     | Integer
    N_DOFLOADED           | Number of DOF forces                                   | Integer
    N_DOFSPRINGS          | Number of DOF spring elements                          | Integer
    COORDINATES           | Coordinates properties                                 | Py Numpy array
                          |    Node, x, y                                          |
    ELEMENTS              | Elements properties                                    | Py Numpy array
                          |    Node 0 ... Node (N_NODES - 1), Material ID,         | 
                          |    Geometry ID, Hinge ID node 0, Hinge ID node 1       |
    MATERIALS             | Materials properties                                   | Py Numpy array
                          |    Young, Poisson, Density, Thermal coefficient        |
    SECTIONS              | Sections properties                                    | Py Numpy array
                          |    Area, Inertia 1, Inertia Frame bar, X GC, Y GC      |
    PRESCRIPTIONS         | Prescribed DOF displacement properties                 | Py Numpy array              
                          |    Node, Direction (X = 0, Y = 1, Z = 2), Value        | 
    NODAL_LOAD            | Nodal DOF force properties                             | Py Numpy array              
                          |    Node, Direction (X = 0, Y = 1, Z = 2), Value        |
    SPRINGS               | Nodal DOF spring properties                            | Py Numpy array              
                          |    Node, Direction (X = 0, Y = 1, Z = 2), Value        |
    """
    # Open file and read dataset
    FILE = open(FILENAME, "r")
    DATASET = FILE.read().split("\n")
    TYPE_ELEMENT = int(DATASET.pop(0).split(':')[1])
    TYPE_SOLUTION = int(DATASET.pop(0).split(':')[1])
    N_NODES = int(DATASET.pop(0).split(':')[1])
    N_MATERIALS = int(DATASET.pop(0).split(':')[1])
    N_SECTIONS = int(DATASET.pop(0).split(':')[1])
    N_ELEMENTS = int(DATASET.pop(0).split(':')[1])
    N_DOFPRESCRIPTIONS = int(DATASET.pop(0).split(':')[1])
    N_DOFLOADED  = int(DATASET.pop(0).split(':')[1]) 
    N_DOFSPRINGS = int(DATASET.pop(0).split(':')[1]) 
    DATASET.pop(0)
    DATASET.pop(0)
    # Coordinates
    COORDINATES = np.zeros((N_NODES, 2))
    for I_COUNT in range(N_NODES):
        VALUES = DATASET.pop(0).split(',')
        COORDINATES[int(VALUES[0]),0] = float(VALUES[1])
        COORDINATES[int(VALUES[0]),1] = float(VALUES[2])
    DATASET.pop(0)
    DATASET.pop(0)
    # Elements
    ELEMENTS = np.zeros((N_ELEMENTS, 6))
    for J_COUNT in range(N_ELEMENTS):
        VALUES = DATASET.pop(0).split(',')
        ELEMENTS[int(VALUES[0]),0] = int(VALUES[1])
        ELEMENTS[int(VALUES[0]),1] = int(VALUES[2])    
        ELEMENTS[int(VALUES[0]),2] = int(VALUES[3])
        ELEMENTS[int(VALUES[0]),3] = int(VALUES[4])
        ELEMENTS[int(VALUES[0]),4] = int(VALUES[5])
        ELEMENTS[int(VALUES[0]),5] = int(VALUES[6])
    DATASET.pop(0)
    DATASET.pop(0)
    # Materials
    MATERIALS = np.zeros((N_MATERIALS, 4))
    for K_COUNT in range(N_MATERIALS):
        VALUES = DATASET.pop(0).split(',')
        MATERIALS[int(VALUES[0]),0] = float(VALUES[1])
        MATERIALS[int(VALUES[0]),1] = float(VALUES[2])    
        MATERIALS[int(VALUES[0]),2] = float(VALUES[3])
        MATERIALS[int(VALUES[0]),3] = float(VALUES[4])
    DATASET.pop(0)
    DATASET.pop(0)
    # Sections
    SECTIONS = np.zeros((N_SECTIONS, 5))
    for L_COUNT in range(N_SECTIONS):
        VALUES = DATASET.pop(0).split(',')
        SECTIONS[int(VALUES[0]),0] = float(VALUES[1])
        SECTIONS[int(VALUES[0]),1] = float(VALUES[2])    
        SECTIONS[int(VALUES[0]),2] = float(VALUES[3])
        SECTIONS[int(VALUES[0]),3] = float(VALUES[4])
        SECTIONS[int(VALUES[0]),4] = float(VALUES[5])
    DATASET.pop(0)
    DATASET.pop(0)
    # Prescribed DOF displacements
    PRESCRIPTIONS = np.zeros((N_DOFPRESCRIPTIONS, 3))
    for M_COUNT in range(N_DOFPRESCRIPTIONS):
        VALUES = DATASET.pop(0).split(',')
        PRESCRIPTIONS[int(VALUES[0]),0] = int(VALUES[1])
        PRESCRIPTIONS[int(VALUES[0]),1] = int(VALUES[2])
        PRESCRIPTIONS[int(VALUES[0]),2] = float(VALUES[3]) 
    DATASET.pop(0)
    DATASET.pop(0)
    """
    **** Under development ****
    ****                   ****
    ****                   ****

    # Element load **** Under development ****
    if N_ELEMENTSLOADED == 0:
        DATASET.pop(0)
        ELEMENT_EXTERNAL_LOAD = "null"
    else:
        ELEMENT_EXTERNAL_LOAD = np.zeros((N_ELEMENTSLOADED, 5))
        for N_COUNT in range(N_ELEMENTSLOADED):
            VALUES = DATASET.pop(0).split(',')
            ELEMENT_EXTERNAL_LOAD[int(VALUES[0]),0] = float(VALUES[1])
            ELEMENT_EXTERNAL_LOAD[int(VALUES[0]),1] = float(VALUES[2])    
            ELEMENT_EXTERNAL_LOAD[int(VALUES[0]),2] = float(VALUES[3])
            ELEMENT_EXTERNAL_LOAD[int(VALUES[0]),3] = float(VALUES[4])
    DATASET.pop(0)
    DATASET.pop(0)

    ****                   ****
    ****                   ****
    **** Under development ****
    """
    # Nodal DOF load
    NODAL_LOAD = np.zeros((N_DOFLOADED, 3))
    for O_COUNT in range(N_DOFLOADED):
        VALUES = DATASET.pop(0).split(',')
        NODAL_LOAD[int(VALUES[0]),0] = float(VALUES[1])
        NODAL_LOAD[int(VALUES[0]),1] = float(VALUES[2]) 
        NODAL_LOAD[int(VALUES[0]),2] = float(VALUES[3]) 
    DATASET.pop(0)
    DATASET.pop(0)
    # Spring DOF elements
    if N_DOFSPRINGS == 0:
        DATASET.pop(0)
        SPRINGS = "null"
    else:
        SPRINGS = np.zeros((N_DOFSPRINGS, 3))
        for P_COUNT in range(N_DOFSPRINGS):
            VALUES = DATASET.pop(0).split(',')
            SPRINGS[int(VALUES[0]),0] = int(VALUES[1])
            SPRINGS[int(VALUES[0]),1] = int(VALUES[2])
            SPRINGS[int(VALUES[0]),2] = float(VALUES[3]) 
    return TYPE_SOLUTION, TYPE_ELEMENT, N_NODES, N_MATERIALS, N_SECTIONS, N_ELEMENTS, N_DOFPRESCRIPTIONS, N_DOFLOADED, N_DOFSPRINGS, COORDINATES, ELEMENTS, MATERIALS, SECTIONS, PRESCRIPTIONS, NODAL_LOAD, SPRINGS

# LEITURA DO ARQUIVO DICIONÁRIO
def GET_VALUE_FROM_DICT_MEF1D_FINITO(DICTIONARY):
    """
    This function reads data from dictionary.

    Input:
    DICTIONARY            | Structural dataset                                     | Py dictionary

    Output: 
    TYPE_ELEMENT          | Type element in Finito algorithm                       | Integer 
                          |     0 - Frame bar element                              |
    TYPE_SOLUTION         | Solution of the system of equations                    | Integer
                          |    0 - Condense procedure                              |
                          |    1 - Zero and One algorithm                          |
    N_NODES               | Number of nodes                                        | Integer
    N_MATERIALS           | Number of materials                                    | Integer
    N_SECTIONS            | Number of sections                                     | Integer
    N_ELEMENTS            | Number of frame elements                               | Integer
    N_DOFPRESCRIPTIONS    | Number of DOF displacement control                     | Integer
    N_DOFLOADED           | Number of DOF forces                                   | Integer
    N_DOFSPRINGS          | Number of DOF spring elements                          | Integer
    COORDINATES           | Coordinates properties                                 | Py Numpy array
                          |    Node, x, y                                          |
    ELEMENTS              | Elements properties                                    | Py Numpy array
                          |    Node 0 ... Node (N_NODES - 1), Material ID,         | 
                          |    Geometry ID, Hinge ID node 0, Hinge ID node 1       |
    MATERIALS             | Materials properties                                   | Py Numpy array
                          |    Young, Poisson, Density, Thermal coefficient        |
    SECTIONS              | Sections properties                                    | Py Numpy array
                          |    Area, Inertia 1, Inertia Frame bar, X GC, Y GC      |
    PRESCRIPTIONS         | Prescribed DOF displacement properties                 | Py Numpy array              
                          |    Node, Direction (X = 0, Y = 1, Z = 2), Value        | 
    NODAL_LOAD            | Nodal DOF force properties                             | Py Numpy array              
                          |    Node, Direction (X = 0, Y = 1, Z = 2), Value        |
    SPRINGS               | Nodal DOF spring properties                            | Py Numpy array              
                          |    Node, Direction (X = 0, Y = 1, Z = 2), Value        |
    """
    # Read dataset
    TYPE_ELEMENT = DICTIONARY["TYPE_ELEMENT"]
    TYPE_SOLUTION = DICTIONARY["TYPE_SOLUTION"]
    N_NODES = DICTIONARY["N_NODES"]
    N_MATERIALS = DICTIONARY["N_MATERIALS"]
    N_SECTIONS = DICTIONARY["N_SECTIONS"]
    N_ELEMENTS = DICTIONARY["N_ELEMENTS"]
    N_DOFPRESCRIPTIONS = DICTIONARY["N_DOFPRESCRIPTIONS"]
    N_DOFLOADED = DICTIONARY["N_DOFLOADED"]
    N_DOFSPRINGS = DICTIONARY["N_DOFSPRINGS"]
    COORDINATES = DICTIONARY["COORDINATES"]
    ELEMENTS = DICTIONARY["ELEMENTS"]
    MATERIALS = DICTIONARY["MATERIALS"]
    SECTIONS = DICTIONARY["SECTIONS"]
    PRESCRIPTIONS = DICTIONARY["PRESCRIBED DISPLACEMENTS"]
    NODAL_LOAD = DICTIONARY["NODAL LOADS"]
    SPRINGS = DICTIONARY["SPRINGS"]
    return TYPE_SOLUTION, TYPE_ELEMENT, N_NODES, N_MATERIALS, N_SECTIONS, N_ELEMENTS, N_DOFPRESCRIPTIONS, N_DOFLOADED, N_DOFSPRINGS, COORDINATES, ELEMENTS, MATERIALS, SECTIONS, PRESCRIPTIONS, NODAL_LOAD, SPRINGS

# ORGANIZAÇÃO DAS PROPRIEDADES DO MATERIAL DO ELEMENTO I 
def MATERIALS_PROPERTIES_0(ELEMENTS, MATERIALS, I_ELEMENT, AUX_1):
    """
    This function creates a vector with the material information of the I_ELEMENT element TYPE_ELEMENT = 0 (Frame element).

    Input:
    ELEMENTS           | Elements properties                                   | Py Numpy array
                       |    ID, Node 0 ... Node (N_NODES - 1), Material ID ,   |
                       |    Geometry ID, Hinge ID node 1, Hinge ID node 2      |
    MATERIALS          | Materials properties                                  | Py Numpy array
                       |     ID, Young, Poisson, Density, Thermal coefficient  |
    I_ELEMENT          | i element in looping                                  | Integer
    AUX_1              | ID material                                           | Integer
    
    Output:
    MATERIAL_IELEMENT  | Material I_ELEMENT properties                         | Py list[5]
                       |     [0] - Young                                       |
                       |     [1] - Shear modulus                               |
                       |     [2] - Poisson                                     |
                       |     [3] - Thermal coefficient                         |
                       |     [4] - Density                                     |
    """
    MATERIAL_ID = int(ELEMENTS[I_ELEMENT, AUX_1])
    E = MATERIALS[MATERIAL_ID, 0]
    NU = MATERIALS[MATERIAL_ID, 1]
    PHO = MATERIALS[MATERIAL_ID, 2]
    ALPHA = MATERIALS[MATERIAL_ID, 3]
    G = E / (2 * (1 + NU))
    MATERIAL_IELEMENT = [E, G, NU, ALPHA, PHO]
    return MATERIAL_IELEMENT

# ORGANIZAÇÃO DAS PROPRIEDADES GEOMÉTRICAS DO ELEMENTO I TIPO FRAME
def GEOMETRIC_PROPERTIES_0(COORDINATES, ELEMENTS, SECTIONS, I_ELEMENT, AUX_2):
    """ 
    This function assigns the bar element's geometric propertiest of the I_ELEMENT element TYPE_ELEMENT = 0 (Frame element).

    Input:
    COORDINATES        | Coordinates properties                                | Py Numpy array
                       |     Node, x, y                                        |
    ELEMENTS           | Elements properties                                   | Py Numpy array
                       |    Node 0 ... Node (N_NODES - 1), Material ID,        |
                       |    Geometry ID, Hinge ID node 1, Hinge ID node 2      | 
    SECTIONS           | Sections properties                                   | Py Numpy array
                       |    Area, Inertia 1, Inertia Frame bar, X GC, Y GC     |
    I_ELEMENT          | i element in looping                                  | Integer
    AUX_2              | ID section                                            | Integer

    Output:
    SECTION_IELEMENT   | Section I_ELEMENT properties                          | Py list[6]
                       |     [0] - Length                                      |
                       |     [1] - Sine                                        |
                       |     [2] - Cosine                                      |
                       |     [3] - Area                                        |
                       |     [4] - Inertia auxiliar                            |
                       |     [5] - Inertia frame element                       |
    """
    NODE_1 = int(ELEMENTS[I_ELEMENT, 0])
    NODE_2 = int(ELEMENTS[I_ELEMENT, 1])
    X_NODE1 = COORDINATES[NODE_1, 0]
    Y_NODE1 = COORDINATES[NODE_1, 1]
    X_NODE2 = COORDINATES[NODE_2, 0]
    Y_NODE2 = COORDINATES[NODE_2, 1]
    DELTA_X = X_NODE2 - X_NODE1
    DELTA_Y = Y_NODE2 - Y_NODE1
    L = ((DELTA_X) ** 2 + (DELTA_Y) ** 2) ** 0.50
    COS = DELTA_X / L
    SIN = DELTA_Y / L
    SECTION_ID = int(ELEMENTS[I_ELEMENT, AUX_2])
    A = SECTIONS[SECTION_ID, 0]
    I_1 = SECTIONS[SECTION_ID, 1]
    I_2 = SECTIONS[SECTION_ID, 2]
    SECTION_IELEMENT = [L, SIN, COS, A, I_1, I_2]
    return SECTION_IELEMENT

# SEPARAÇÃO DE UMA MATRIZ QUE CONTÉM AS PROPRIEDADES DE RÓTULAS
def HINGED_PROPERTIES(ELEMENTS):
    """
    This function creates an array with the hinge properties per node.

    Input
    ELEMENTS  | Elements properties                                | Py Numpy array
              |     Node 0 ... Node (N_NODES - 1), Material ID,    | 
              |     Geometry ID, Hinge ID node 1, Hinge ID node 2  |
    
    Output:
    HINGES    | Hinge properties per node                          | Py Numpy array[N_NODES x 2]
              |     0 - No hinge                                   |
              |     1 - Yes hinge                                  |
    """
    HINGES = ELEMENTS[:, 4:]
    return HINGES

# MATRIZ DE RIGIDEZ LOCAL DO ELEMENTO TIPO FRAME
def ELEMENT_STIFFNESS_0(TYPE_ELEMENT, SECTION_IELEMENT, MATERIAL_IELEMENT, HINGES_IELEMENT):
    """ 
    This function creates the element stiffness matrix of the I_ELEMENT. 
    
    Input:
    TYPE_ELEMENT       | Type element in Finito algorithm                      | Integer 
                       |     0 - Frame bar element                             |        
    SECTION_IELEMENT   | Section I_ELEMENT properties                          | Py list[6]
                       |     [0] - Length                                      |
                       |     [1] - Sine                                        |
                       |     [2] - Cosine                                      |
                       |     [3] - Area                                        |
                       |     [4] - Inertia auxiliar                            |
                       |     [5] - Inertia frame element                       |
    MATERIAL_IELEMENT  | Material I_ELEMENT properties                         | Py list[5]
                       |     [0] - Young                                       |
                       |     [1] - Shear modulus                               |
                       |     [2] - Poisson                                     |
                       |     [3] - Thermal coefficient                         |
                       |     [4] - Density                                     |   
    HINGES             | Hinge properties per node                             | Py Numpy array[N_NODES x 2]
                       |     0 - No hinge                                      |
                       |     1 - Yes hinge                                     |
    
    Output:
    K_IELEMENT         | Local stiffness matrix I_ELEMENT                      | Py Numpy array [N_DOFSELEMENT x N_DOFSELEMENT]
    """
    if (TYPE_ELEMENT == 0 and HINGES_IELEMENT[0] == 0 and HINGES_IELEMENT[1] == 0):
        L = SECTION_IELEMENT[0]
        A = SECTION_IELEMENT[3]
        I = SECTION_IELEMENT[5]
        E = MATERIAL_IELEMENT[0]
        C1 = A * E / L
        C2 = E * I / (L ** 3)
        K_IELEMENT = np.array([[C1, 0, 0, -C1, 0, 0],
                               [0, 12 * C2, 6 * C2 * L, 0, -12 * C2, 6 * C2 * L],
                               [0, 6 * C2 * L, 4 * C2 * L ** 2, 0, -6 * C2 * L, 2 * C2 * L ** 2],
                               [-C1, 0, 0, C1, 0, 0],
                               [0, -12 * C2, -6 * C2 * L, 0, 12 * C2, -6 * C2 * L],
                               [0, 6 * C2 * L, 2 * C2 * L ** 2, 0, -6 * C2 * L, 4 * C2 * L **2]])
    elif (TYPE_ELEMENT == 0 and HINGES_IELEMENT[0] == 0 and HINGES_IELEMENT[1] == 1):
        L = SECTION_IELEMENT[0]
        A = SECTION_IELEMENT[3]
        I = SECTION_IELEMENT[5]
        E = MATERIAL_IELEMENT[0]
        C1 = A * E / L
        C2 = E * I / (L ** 3)
        K_IELEMENT = np.array([[C1, 0, 0, -C1, 0, 0],
                               [0, 3 * C2, 3 * C2 * L, 0, -3 * C2, 0],
                               [0, 3 * C2 * L, 3 * C2 * L ** 2, 0, -3 * C2 * L, 0],
                               [-C1, 0, 0, C1, 0, 0],
                               [0, -3 * C2, -6 * C2 * L, 0, 3 * C2, 0],
                               [0, 0, 0, 0, 0, 0]])    
    elif (TYPE_ELEMENT == 0 and HINGES_IELEMENT[0] == 1 and HINGES_IELEMENT[1] == 0):
        L = SECTION_IELEMENT[0]
        A = SECTION_IELEMENT[3]
        I = SECTION_IELEMENT[5]
        E = MATERIAL_IELEMENT[0]
        C1 = A * E / L
        C2 = E * I / (L ** 3)
        K_IELEMENT = np.array([[C1, 0, 0, -C1, 0, 0],
                               [0, 3 * C2, 0, 0, -3 * C2, 3 * C2 * L],
                               [0, 0, 0, 0, 0, 0],
                               [-C1, 0, 0, C1, 0, 0],
                               [0, -3 * C2, 0, 0, 3 * C2, -3 * C2 * L],
                               [0, 3 * C2 * L, 0, 0, -3 * C2 * L, 3 * C2 * L **2]])     
    elif (TYPE_ELEMENT == 0 and HINGES_IELEMENT[0] == 1 and HINGES_IELEMENT[1] == 1):
        L = SECTION_IELEMENT[0]
        A = SECTION_IELEMENT[3]
        I = SECTION_IELEMENT[5]
        E = MATERIAL_IELEMENT[0]
        C1 = A * E / L
        C2 = E * I / (L ** 3)
        K_IELEMENT = np.array([[C1, 0, 0, -C1, 0, 0],
                              [0, 0, 0, 0, 0, 0],
                              [0, 0, 0, 0, 0, 0],
                              [-C1, 0, 0, C1, 0, 0],
                              [0, 0, 0, 0, 0, 0],
                              [0, 0, 0, 0, 0, 0]])
    return K_IELEMENT

def ELEMENT_ROTATION(TYPE_ELEMENT, SECTION_IELEMENT):
    """ 
    This function creates the rotation matrix of the I_ELEMENT element.

    Input:
    TYPE_ELEMENT       | Type element in Finito algorithm                      | Integer 
                       |     0 - Frame bar element                             |
    SECTION_IELEMENT   | Section I_ELEMENT properties                          | Py list[6]
                       |     [0] - Length                                      |
                       |     [1] - Sine                                        |
                       |     [2] - Cosine                                      |
                       |     [3] - Area                                        |
                       |     [4] - Inertia auxiliar                            |
                       |     [5] - Inertia frame element                       |  
    Output:
    R_IELEMENT         | Rotation matrix  I_ELEMENT                            | Py Numpy array [N_DOFSELEMENT x N_DOFSELEMENT]
    """
    SIN = SECTION_IELEMENT[1]
    COS = SECTION_IELEMENT[2]
    if TYPE_ELEMENT == 0:
        R_IELEMENT = np.array([[COS, SIN, 0, 0, 0, 0],
                               [-SIN, COS, 0, 0, 0, 0],
                               [0, 0, 1, 0, 0, 0],
                               [0, 0, 0, COS, SIN, 0],
                               [0, 0, 0, -SIN, COS, 0],
                               [0, 0, 0, 0, 0, 1]])
    return R_IELEMENT                                                                                                                                                                                                                                                                                                                                                             
                                                                                                                                                                    
#   /$$$$$$  /$$$$$$$  /$$$$$$$$ /$$$$$$$$       /$$$$$$$$ /$$$$$$$$  /$$$$$$  /$$   /$$ /$$   /$$  /$$$$$$  /$$        /$$$$$$   /$$$$$$  /$$$$$$ /$$$$$$$$  /$$$$$$ 
#  /$$__  $$| $$__  $$| $$_____/| $$_____/      |__  $$__/| $$_____/ /$$__  $$| $$  | $$| $$$ | $$ /$$__  $$| $$       /$$__  $$ /$$__  $$|_  $$_/| $$_____/ /$$__  $$
# | $$  \__/| $$  \ $$| $$      | $$               | $$   | $$      | $$  \__/| $$  | $$| $$$$| $$| $$  \ $$| $$      | $$  \ $$| $$  \__/  | $$  | $$      | $$  \__/
# | $$ /$$$$| $$$$$$$/| $$$$$   | $$$$$            | $$   | $$$$$   | $$      | $$$$$$$$| $$ $$ $$| $$  | $$| $$      | $$  | $$| $$ /$$$$  | $$  | $$$$$   |  $$$$$$ 
# | $$|_  $$| $$____/ | $$__/   | $$__/            | $$   | $$__/   | $$      | $$__  $$| $$  $$$$| $$  | $$| $$      | $$  | $$| $$|_  $$  | $$  | $$__/    \____  $$
# | $$  \ $$| $$      | $$      | $$               | $$   | $$      | $$    $$| $$  | $$| $$\  $$$| $$  | $$| $$      | $$  | $$| $$  \ $$  | $$  | $$       /$$  \ $$
# |  $$$$$$/| $$      | $$$$$$$$| $$$$$$$$         | $$   | $$$$$$$$|  $$$$$$/| $$  | $$| $$ \  $$|  $$$$$$/| $$$$$$$$|  $$$$$$/|  $$$$$$/ /$$$$$$| $$$$$$$$|  $$$$$$/
#  \______/ |__/      |________/|________/         |__/   |________/ \______/ |__/  |__/|__/  \__/ \______/ |________/ \______/  \______/ |______/|________/ \______/ 


"""
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░▒▒▒▒▒░░░░░░░░░░▒▒▒▒░░░░░░░░░░░▒▒▒░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░▒▓▓▓▓▓▓▓▒░░░░░░▒▒▒▒▒▒▒▒░░░░░░▒▒▒▒▒▒▒▒░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░▓▓▓▓▓▓▓▓▓░░░░░░▒▒▒▒▒▒▒▒░░░░░░▒▒▒▒▒▒▒▒▒░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░▓▓▓▓▓▓▓░░░░░░░▒▒▒▒▒▒▒░░░░░░░░▒▒▒▒▒▒▒░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░▒▒▒░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░██▓▓▓▓▒░░░░░░░▒█▒░░░░░░░░▓█▓░░░░██░░░░░░░░░█▒░░░░░░░▒▓▓▓█▓▓▓░░░░░░░░▒▓▓▓▓▓▓▓░░░░░░
░░░░░█▓░░░░░░░░░░░░▒█▒░░░░░░░░▓█▓█▒░░██░░░░░░░░░█▒░░░░░░░░░░▒█░░░░░░░░░░▒█▒░░░░░▒█▒░░░░
░░░░░██▓▓▓▓░░░░░░░░▒█▒░░░░░░░░▓▓░▒█▓░██░░░░░░░░░█▒░░░░░░░░░░▒█░░░░░░░░░░▓█░░░░░░░█▓░░░░
░░░░░█▓░░░░░░░░░░░░▒█▒░░░░░░░░▓▓░░░▓▓██░░░░░░░░░█▒░░░░░░░░░░▒█░░░░░░░░░░▒█▓░░░░░▓█░░░░░
░░░░░▓▓░░░░░░░░░░░░▒▓░░░░░░░░░▓▓░░░░▒▓▓░░░░░░░░░▓▒░░░░░░░░░░▒▓░░░░░░░░░░░░▒▓▓▓▓▓▒░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
"""
################################################################################
# UNIVERSIDADE FEDERAL DE CATALÃO (UFCAT)
# WANDERLEI MALAQUIAS PEREIRA JUNIOR,                  ENG. CIVIL / PROF (UFCAT)
# MARCOS NAPOLEÃO RABELO,                              ENG. CIVIL / PROF (UFCAT)
# DAVIDSON DE OLIVEIRA FRANÇA JUNIOR,                    ENG. CIVIL / PROF (UNA)
# GABRIEL BERNARDES CARVALHO,                                 ENG. CIVIL (UFCAT)
# JOSÉ VITOR CARVALHO SILVA,                                  ENG. CIVIL (UFCAT)
# MURILO CARNEIRO RODRIGUES,                                  ENG. CIVIL (UFCAT)
################################################################################

################################################################################
# DESCRIÇÃO ALGORITMO:
# BIBLIOTECA FINITO PARA FUNÇÕES DO ALGORITMO MEF2D DESENVOLVIDA PELO GRUPO DE 
# PESQUISA E ESTUDOS EM ENGENHARIA (GPEE)
################################################################################

################################################################################
# BIBLIOTECAS NATIVAS PYTHON
import numpy as np

################################################################################
# BIBLIOTECAS DESENVOLVEDORES GPEE
    
def GET_VALUE_FROM_DICT_MEF2D_FINITO(DICTIONARY):
    """
    This function read input data from dictionary.

    Input:
    FILENAME             | Structural dataset                                     | Py dictionary

    Output: 
    N_NODES              | Number of nodes                                        | Integer
    N_MATERIALS          | Number of materials                                    | Integer
    N_THICKNESS          | Number of thicness                                     | Integer
    N_ELEMENTS           | Number of CST element                                  | Integer
    N_FORCES             | Number of nodal forces                                 | Integer
    N_PRESSURES          | Number of element pressures                            | Integer
    N_DISPLACEMENTS      | Number of nodal displacement control                   | Integer
    TYPE_PLANE           | Type of analysis in the plan                           | String
                         |      'EPT' - Plane Stress                              |
                         |      'EPD' - Plane Strain                              |
    TYPE_ELEMENT         | Type element in Finito algorithm                       | Integer 
                         |      0 - Frame bar element                             |
                         |      1 - CST surface element                           |
    TYPE_SOLUTION        | Solution of the system of equations                    | Integer
                         |      0 - Condense procedure                            |
                         |      1 - 0 and 1 algorithm                             |
    TYPE_INTEGRATION     | Type numerical integration                             | String
                         |      1 - Hammer 12 points integration                  |
    COORDINATES          | Coordinates properties                                 | Py Numpy array
                         |      Node, X, Y                                        |
    MATERIALS            | Materials properties                                   | Py Numpy array
                         |      ID, Young, Poisson, Density                       |
    THICKNESS            | Thickness properties                                   | Py Numpy array
                         |      ID, Thickness                                     |
    ELEMENTS             | Elements properties                                    | Py Numpy array
                         |      ID, Node 0 ... Node (N_DODES - 1), Material ID,   |
                         |      Thickness ID                                      |
    NODAL_EXTERNAL_LOAD  | Nodal force properties                                 | Py Numpy array              
                               ID, node ID, FX value, FY value 
    # # # # # # # # # # # # # # # # # # # # # # # # #

    PRESSURES: Under development 
    
    # # # # # # # # # # # # # # # # # # # # # # # # #
    PRESCRIPTIONS:       | Displacement properties                                | Py Numpy array  
                         |  ID, Node ID, Direction ('X', Y' and 'BOTH'), Displa-  |
                         |  cement value                                          |
    """
    N_NODES = DICTIONARY["N_NODES"]
    N_MATERIALS = DICTIONARY["N_MATERIALS"]
    N_THICKNESS = DICTIONARY["N_THICKNESS"]
    N_ELEMENTS = DICTIONARY["N_ELEMENTST3"]
    N_ELEMENTST6 = DICTIONARY["N_ELEMENTST6"]
    N_ELEMENTST10 = DICTIONARY["N_ELEMENTST10"]
    N_FORCES = DICTIONARY["N_FORCES"]
    N_PRESSURES = DICTIONARY["N_PRESSURES"]
    N_DISPLACEMENTS = DICTIONARY["N_DISCPLACEMENTS"]
    TYPE_PLANE = DICTIONARY["TYPE_PLANE"]
    TYPE_ELEMENT = DICTIONARY["TYPE_ELEMENT"]
    TYPE_SOLUTION = DICTIONARY["TYPE_SOLUTION"]
    GRAU_INT = DICTIONARY["GRAU_INT"]
    COORDINATES = DICTIONARY["COORDINATES"]
    MATERIALS = DICTIONARY["MATERIALS"]
    THICKNESS = DICTIONARY["THICKNESS"]
    ELEMENTS = DICTIONARY["ELEMENTS"]
    NODAL_EXTERNAL_LOAD = DICTIONARY["EXTERNAL LOADS"]
    PRESCRIPTIONS = DICTIONARY["PRESCRIBED DISPLACEMENTS"]
    return N_NODES, N_MATERIALS, N_THICKNESS, N_ELEMENTS, N_ELEMENTST6, N_ELEMENTST10, N_FORCES, N_PRESSURES, N_DISPLACEMENTS, N_DISPLACEMENTS, TYPE_PLANE, TYPE_ELEMENT, TYPE_SOLUTION, GRAU_INT, COORDINATES, MATERIALS, THICKNESS, ELEMENTS, NODAL_EXTERNAL_LOAD, PRESCRIPTIONS

def GET_VALUE_FROM_TXT_MEF2D_FINITO(FILENAME):
    """
    This function read input data from .txt file.

    Input:
    FILENAME             | Structural dataset                                     | .txt extension

    Output: 
    N_NODES              | Number of nodes                                        | Integer
    N_MATERIALS          | Number of materials                                    | Integer
    N_THICKNESS          | Number of thicness                                     | Integer
    N_ELEMENTS           | Number of CST element                                  | Integer
    N_FORCES             | Number of nodal forces                                 | Integer
    N_PRESSURES          | Number of element pressures                            | Integer
    N_DISPLACEMENTS      | Number of nodal displacement control                   | Integer
    TYPE_PLANE           | Type of analysis in the plan                           | String
                         |      'EPT' - Plane Stress                              |
                         |      'EPD' - Plane Strain                              |
    TYPE_ELEMENT         | Type element in Finito algorithm                       | Integer 
                         |      0 - Frame bar element                             |
                         |      1 - CST surface element                           |
    TYPE_SOLUTION        | Solution of the system of equations                    | Integer
                         |      0 - Condense procedure                            |
                         |      1 - 0 and 1 algorithm                             |
    TYPE_INTEGRATION     | Type numerical integration                             | String
                         |      1 - Hammer 12 points integration                  |
    COORDINATES          | Coordinates properties                                 | Py Numpy array
                         |      Node, X, Y                                        |
    MATERIALS            | Materials properties                                   | Py Numpy array
                         |      ID, Young, Poisson, Density                       |
    THICKNESS            | Thickness properties                                   | Py Numpy array
                         |      ID, Thickness                                     |
    ELEMENTS             | Elements properties                                    | Py Numpy array
                         |      ID, Node 0 ... Node (N_DODES - 1), Material ID,   |
                         |      Thickness ID                                      |
    NODAL_EXTERNAL_LOAD  | Nodal force properties                                 | Py Numpy array              
                               ID, node ID, FX value, FY value 
    # # # # # # # # # # # # # # # # # # # # # # # # #

    PRESSURES: Under development 
    
    # # # # # # # # # # # # # # # # # # # # # # # # #
    PRESCRIPTIONS:       | Displacement properties                                | Py Numpy array  
                         |  ID, Node ID, Direction ('X', Y' and 'BOTH'), Displa-  |
                         |  cement value                                          |
    """
    # Read input file general properties
    FILE = open(FILENAME, "r")
    DATASET = FILE.read().split("\n")
    N_NODES = int(DATASET.pop(0).split(':')[1])
    N_MATERIALS = int(DATASET.pop(0).split(':')[1])
    N_THICKNESS = int(DATASET.pop(0).split(':')[1])
    N_ELEMENTS = int(DATASET.pop(0).split(':')[1])
    N_ELEMENTST6 = int(DATASET.pop(0).split(':')[1])
    N_ELEMENTST10 = int(DATASET.pop(0).split(':')[1])
    N_FORCES = int(DATASET.pop(0).split(':')[1])
    N_PRESSURES = int(DATASET.pop(0).split(':')[1])
    N_DISPLACEMENTS = int(DATASET.pop(0).split(':')[1])
    TYPE_PLANE = int(DATASET.pop(0).split(':')[1])
    TYPE_ELEMENT = int(DATASET.pop(0).split(':')[1])
    TYPE_SOLUTION = int(DATASET.pop(0).split(':')[1])
    TYPE_INTEGRATION = int(DATASET.pop(0).split(':')[1])
    # Read coordinates 
    DATASET.pop(0)
    DATASET.pop(0)   
    COORDINATES = np.zeros((N_NODES, 2))
    for I_COUNT in range(N_NODES):
        VALUES = DATASET.pop(0).split(',')
        COORDINATES[int(VALUES[0]), 0] = float(VALUES[1])
        COORDINATES[int(VALUES[0]), 1] = float(VALUES[2])
    # Read materials
    DATASET.pop(0)
    DATASET.pop(0)
    MATERIALS = np.zeros((N_MATERIALS, 3))
    for J_COUNT in range(N_MATERIALS):
        VALUES1 = DATASET.pop(0).split(',')
        MATERIALS[int(VALUES1[0]), 0] = float(VALUES1[1])
        MATERIALS[int(VALUES1[0]), 1] = float(VALUES1[2])
    # Read thickness
    DATASET.pop(0)
    DATASET.pop(0)
    THICKNESS = np.zeros((N_THICKNESS, 1))
    for K_COUNT in range(N_THICKNESS):
        VALUES2 = DATASET.pop(0).split(',')
        THICKNESS[int(VALUES2[0]), 0] = float(VALUES2[1])
    # Read Elements
    DATASET.pop(0)
    DATASET.pop(0)
    # COLOCAR AQUI UM IF PARA DIZER QUE ESSA LEITURA É SÓ DO T3
    ELEMENTS = np.zeros((N_ELEMENTS, 5))
    for L_COUNT in range(N_ELEMENTS):
        VALUES3 = DATASET.pop(0).split(',')
        ELEMENTS[int(VALUES3[0]), 0] = int(VALUES3[1])
        ELEMENTS[int(VALUES3[0]), 1] = int(VALUES3[2])
        ELEMENTS[int(VALUES3[0]), 2] = int(VALUES3[3])
        ELEMENTS[int(VALUES3[0]), 3] = int(VALUES3[4])
        ELEMENTS[int(VALUES3[0]), 4] = int(VALUES3[5])
    # Read nodal forces
    DATASET.pop(0)
    DATASET.pop(0)
    NODAL_EXTERNAL_LOAD = np.zeros((N_FORCES, 3))
    for M_COUNT in range(N_FORCES):
        VALUES4 = DATASET.pop(0).split(',')
        NODAL_EXTERNAL_LOAD[int(VALUES4[0]), 0] = float(VALUES4[1])  # Nó
        if float(VALUES4[2]) != 0:
            NODAL_EXTERNAL_LOAD[int(VALUES4[0]), 2] = float(VALUES4[2])
            NODAL_EXTERNAL_LOAD[int(VALUES4[0]), 1] = 0
        elif float(VALUES4[3]) != 0:
            NODAL_EXTERNAL_LOAD[int(VALUES4[0]), 2] = float(VALUES4[3])
            NODAL_EXTERNAL_LOAD[int(VALUES4[0]), 1] = 1
    # # # # # # # # # # # # # # # # # # # # # # # # #
    
    # Under development Pressure read
    
    # # # # # # # # # # # # # # # # # # # # # # # # # 
    # Read displacement
    DATASET.pop(0)
    DATASET.pop(0)
    COUNT = 0
    for K_COUNT in range(N_DISPLACEMENTS):
        VALUES5 = DATASET[K_COUNT].split(',')

        if VALUES5[2] == "'BOTH'":
            COUNT += 2

        else:
            COUNT += 1
    PRESCRIPTIONS = np.zeros((COUNT, 3))
    COUNT = 0
    for w in range(N_DISPLACEMENTS):
        VALUES6 = DATASET.pop(0).split(',')
        if VALUES6[2] == "'X'":
            PRESCRIPTIONS[COUNT, 0] = int(VALUES6[1])
            PRESCRIPTIONS[COUNT, 1] = 0
            PRESCRIPTIONS[COUNT, 2] = float(VALUES6[3])
            COUNT += 1
        elif VALUES6[2] == "'Y'":
            PRESCRIPTIONS[COUNT, 0] = int(VALUES6[1])
            PRESCRIPTIONS[COUNT, 1] = 1
            PRESCRIPTIONS[COUNT, 2] = float(VALUES6[3])
            COUNT += 1
        else:
            PRESCRIPTIONS[COUNT, 0] = int(VALUES6[1])
            PRESCRIPTIONS[COUNT, 1] = 0
            PRESCRIPTIONS[COUNT, 2] = float(VALUES6[3])
            COUNT += 1
            PRESCRIPTIONS[COUNT, 0] = int(VALUES6[1])
            PRESCRIPTIONS[COUNT, 1] = 1
            PRESCRIPTIONS[COUNT, 2] = float(VALUES6[3])
            COUNT += 1
    return N_NODES, N_MATERIALS, N_THICKNESS, N_ELEMENTS, N_ELEMENTST6, N_ELEMENTST10, N_FORCES, N_PRESSURES, N_DISPLACEMENTS, TYPE_PLANE, TYPE_ELEMENT, TYPE_SOLUTION, TYPE_INTEGRATION, COORDINATES, MATERIALS, THICKNESS, ELEMENTS, NODAL_EXTERNAL_LOAD, PRESCRIPTIONS

# ORGANIZAÇÃO DAS PROPRIEDADES GEOMÉTRICAS DO ELEMENTO I TIPO PLANO
def GEOMETRIC_PROPERTIES_1(COORDINATES, ELEMENTS, I_ELEMENT, THICKNESS, AUX_2):
    """ 
    This function assigns the surface element's geometric propertiest of the I_ELEMENT element TYPE_ELEMENT = 1 (CST element).
    
    Input:
    COORDINATES        | Coordinates properties                                | Py Numpy array
                       |    Node, x, y                                         |
    ELEMENTS           | Elements properties                                   | Py Numpy array
                       |    Node 0 ... Node (N_NODES - 1), Material ID,        |
                       |    Thickness ID                                       |
    THICKNESS          | Thickness properties                                  | Py Numpy array
                       |    Thickness                                          |
    I_ELEMENT          | ID i element in looping for                           | Integer
    AUX_2              | ID thickness                                          | Integer

    Output: 
    SECTION_IELEMENT   | Section I_ELEMENT properties                          | Py dictionary
                       |     ['X_E']       - Elements coordinates              |                                   
                       |     ['THICKNESS'] - Sine                              |
    """
    NODE_0 = int(ELEMENTS[I_ELEMENT, 0])
    X0, Y0 = COORDINATES[NODE_0, 0], COORDINATES[NODE_0, 1]
    NODE_1 = int(ELEMENTS[I_ELEMENT, 1])
    X1, Y1 = COORDINATES[NODE_1, 0], COORDINATES[NODE_1, 1]
    NODE_2 = int(ELEMENTS[I_ELEMENT, 2])
    X2, Y2 = COORDINATES[NODE_2, 0], COORDINATES[NODE_2, 1]
    THICK_ID = int(ELEMENTS[I_ELEMENT, AUX_2])
    THICK = THICKNESS[THICK_ID, 0]
    SECTION_IELEMENT = {'X_E': np.array([[X0, Y0], [X1, Y1], [X2, Y2]]), 'THICKNESS': THICK}
    return SECTION_IELEMENT

# MATRIZ CONSTITUTIVA DO ELEMENTO TIPO PLANO
def CONSTITUTIVE_C(TYPE_PLANE, MATERIALS, ELEMENTS, I_ELEMENT):
    """
    This function determines the matrix responsible for establishing the constitutive relationship between stress and strain according to the choosen analysis.

    Input:
    TYPE_PLANE  | Type of analysis in the plan                           | String
                |      'EPT' - Plane Stress                              |
                |      'EPD' - Plane Strain                              |
    MATERIALS   | Materials properties                                   | Py Numpy array
                |      ID, Young, Poisson, Density                       |
    ELEMENTS    | Elements properties                                    | Py Numpy array
                |    Node 0 ... Node (N_NODES - 1), Material ID,         |
                |    Thickness ID                                        |
    I_ELEMENT   | ID i element in looping for                            | Integer

    Output:
    C           | Constitutive matrix                                    | Py Numpy array[3 x 3]
    """
    MATERIAL_ID = int(ELEMENTS[I_ELEMENT, 4])
    E = MATERIALS[MATERIAL_ID, 0]
    NU = MATERIALS[MATERIAL_ID, 1]
    # Plane stress
    if TYPE_PLANE == 0:
        C11 = 1
        C12 = NU
        C21 = C12
        C22 = 1
        C33 = 0.5 * (1 - NU)
        AUX_1 = E / (1 - NU ** 2)
        AUX_2 = np.array([[C11, C12, 0], [C21, C22, 0], [0, 0, C33]])
    # Plane strain
    elif TYPE_PLANE == 1:
        C11 = 1 - NU
        C12 = NU
        C21 = NU
        C22 = 1 - NU
        C33 = 0.5 - NU
        AUX_1 = E/((1 + NU)*(1 - 2*NU))
        AUX_2 = np.array([[C11, C12, 0],[C21, C22, 0],[0, 0, C33]])
    C = AUX_1 * AUX_2
    return C

def SHAPE_FUNCTIONS(TYPE_ELEMENT, N_NODESELEMENT, ISO_COORDINATES):
    """
    This function creates the matrix of the derivatives of the shape functions

    Input:
    TYPE_ELEMENT         | Type element in Finito algorithm                       | Integer 
                         |      0 - Frame bar element                             |
                         |      1 - CST surface element                           |
    N_NODESELEMENT: Number of nodes per element (integer);
    ISO_COORDINATES: Isoparametric coordinates (Python dictionary);

    Output:
    ND_DIFF: ND derivatives matrix (Python Numpy array);
    NX_DIFF: NX derivatives matrix (Python Numpy array);
    """
    if TYPE_ELEMENT == 1:
        # Derivative shape functions 
        DIFF_KSI = [-1, 1, 0]
        DIFF_ETA = [-1, 0, 1]
        # ND e NX assembly
        NX_DIFF = np.array([DIFF_KSI, DIFF_ETA])
        ND_DIFF = DND_ASSEMBLY(N_NODESELEMENT, NX_DIFF)
    return ND_DIFF, NX_DIFF

def DND_ASSEMBLY(N_NODESELEMENT, NX_DIFF):
    """
    This function assembles the derived matrix ND

    Input:
    N_NODESELEMENT: Number of nodes per element (integer);
    NX_DIFF: NX derivatives matrix (Python Numpy array);
    
    Output:
    ND_DIFF: ND derivatives matrix (Python Numpy array);
    """
    ND_DIFF_1 = np.zeros((2, 2 * N_NODESELEMENT))
    ND_DIFF_2 = np.zeros((2, 2 * N_NODESELEMENT))
    # Automatic assembly 
    for I_COUNT in range(2):
        COUNT_1 = 0
        COUNT_2 = 0
        for J_COUNT in range(0, 2 * N_NODESELEMENT, 2):
            ND_DIFF_1[I_COUNT, J_COUNT] = NX_DIFF[I_COUNT, COUNT_1]
            COUNT_1 += 1
        for K_COUNT in range(1, 2 * N_NODESELEMENT, 2):
            ND_DIFF_2[I_COUNT, K_COUNT] = NX_DIFF[I_COUNT, COUNT_2]
            COUNT_2 += 1
    ND_DIFF = np.vstack((ND_DIFF_1, ND_DIFF_2))
    return ND_DIFF

def STIFFNESS(NX_DIFF, ND_DIFF, C, X_E, THICK):
    """
    This function assembles the element's stiffness matrix to a Gaussian point

    Input:
    NX_DIFF: NX derivatives matrix (Python Numpy array);
    ND_DIFF: ND derivatives matrix (Python Numpy array);
    C: Constitutive matrix in formulation (Python Numpy array); 
    X_E: i element coordinates (Python Numpy array);

    Output:
    K_I: i element stiffness matrix (Python Numpy array);
    """

    # Jacobian matrix
    J = np.dot(NX_DIFF, X_E)
    # Determinant of the Jacobian matrix
    DET_J = np.linalg.det(J)
    # Gamma and Gamma U matrix
    GAMMA = np.linalg.inv(J)
    GAMMA_00 = GAMMA[0, 0]
    GAMMA_01 = GAMMA[0, 1]
    GAMMA_10 = GAMMA[1, 0]
    GAMMA_11 = GAMMA[1, 1]
    GAMMA_U = np.array([[GAMMA_00, GAMMA_01, 0, 0],
                        [GAMMA_10, GAMMA_11, 0, 0],
                        [0, 0, GAMMA_00, GAMMA_01],
                        [0, 0, GAMMA_10, GAMMA_11]])
    H = np.array([[1, 0, 0, 0],
                    [0, 0, 0, 1],
                    [0, 1, 1, 0]])
    B = np.dot(np.dot(H, GAMMA_U), ND_DIFF)
    # i element stiffness matrix
    K_I = THICK * (np.dot(np.dot(B.transpose(), C), np.dot(B, DET_J)))
    return K_I

def NUMERICAL_INTEGRATION(TYPE_INTEGRATION):
    """
    This function creates the parameters for numerical integration
    
    Input:
    TYPE_INTEGRATION: Type numerical integration (string);

    Output:
    NUM_INT: Setup numerical integration (Python dictionary);
    """
    # Hammer 12 points
    # if TYPE_INTEGRATION == 'HAMMER-12':
    if TYPE_INTEGRATION == 1:
        W = [0.116786275726379/2.0, 0.116786275726379/2.0, 0.116786275726379/2.0,
                            0.050844906370207/2.0, 0.050844906370207/2.0, 0.050844906370207/2.0,
                            0.082851075618374/2.0, 0.082851075618374/2.0, 0.082851075618374/2.0,
                            0.082851075618374/2.0, 0.082851075618374/2.0, 0.082851075618374/2.0]
        KSI = [0.501426509658179, 0.249286745170910, 0.249286745170910,
                               0.873821971016996, 0.063089014491502, 0.063089014491502,
                               0.053145049844816, 0.310352451033785, 0.636502499121399,
                               0.310352451033785, 0.636502499121399, 0.053145049844816]
        ETA = [0.249286745170910, 0.249286745170910, 0.501426509658179,
                               0.063089014491502, 0.063089014491502, 0.873821971016996,
                               0.310352451033785, 0.636502499121399, 0.053145049844816,
                               0.053145049844816, 0.310352451033785, 0.636502499121399]
        # N_POINTS = W.shape[1] - shape não funciona em lista
        N_POINTS = len(W)
    NUM_INT = {'W': W, 'KSI': KSI, 'ETA': ETA, 'N': N_POINTS}
    return NUM_INT

def ELEMENT_STIFFNESS_1(NUM_INT, N_DOFSELEMENT, TYPE_ELEMENT, N_NODESELEMENT, C_IELEMENT, SECTION_IELEMENT):
    """
    This function calculates the stiffness matrix of the isoparametric element

    Input:
    NUM_INT: Setup numerical integration (Python dictionary);
    N_DOFSELEMENT: Number of degress of freedom per element (integer);
    TYPE_ELEMENT: Type element in Finito algorithm (integer); 
            0 - Frame bar element
            1 - CST surface element
    N_NODESELEMENT: Number of nodes per element (integer);
    C_IELEMENT: i element constitutive matrix in formulation (Python Numpy array); 
    SECTION_IELEMENT: i element geometric properties (Python dictionary);

    Output:
    K_IELEMENT: Complete i element stiffness matrix
    """
    POINTS = NUM_INT['N']
    K_IELEMENT = np.zeros((N_DOFSELEMENT, N_DOFSELEMENT))
    for I_COUNT in range(POINTS):
        ISO_COORDINATES = {'KSI': NUM_INT['KSI'][I_COUNT], 'ETA': NUM_INT['ETA'][I_COUNT]}
        [ND_DIFF, NX_DIFF] = SHAPE_FUNCTIONS(TYPE_ELEMENT, N_NODESELEMENT, ISO_COORDINATES)
        X_E = SECTION_IELEMENT['X_E']
        THICK = SECTION_IELEMENT['THICKNESS']
        K_I = STIFFNESS(NX_DIFF, ND_DIFF, C_IELEMENT, X_E, THICK)
        WEIGHT = NUM_INT['W'][I_COUNT]
        K_IELEMENT += K_I * WEIGHT
    return K_IELEMENT

#  /$$$$$$$$ /$$$$$$ /$$   /$$ /$$$$$$ /$$$$$$$$  /$$$$$$        /$$$$$$$$  /$$$$$$   /$$$$$$  /$$       /$$$$$$$   /$$$$$$  /$$   /$$                                
# | $$_____/|_  $$_/| $$$ | $$|_  $$_/|__  $$__/ /$$__  $$      |__  $$__/ /$$__  $$ /$$__  $$| $$      | $$__  $$ /$$__  $$| $$  / $$                                
# | $$        | $$  | $$$$| $$  | $$     | $$   | $$  \ $$         | $$   | $$  \ $$| $$  \ $$| $$      | $$  \ $$| $$  \ $$|  $$/ $$/                                
# | $$$$$     | $$  | $$ $$ $$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$$$$$$ | $$  | $$ \  $$$$/                                 
# | $$__/     | $$  | $$  $$$$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$__  $$| $$  | $$  >$$  $$                                 
# | $$        | $$  | $$\  $$$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$  \ $$| $$  | $$ /$$/\  $$                                
# | $$       /$$$$$$| $$ \  $$ /$$$$$$   | $$   |  $$$$$$/         | $$   |  $$$$$$/|  $$$$$$/| $$$$$$$$| $$$$$$$/|  $$$$$$/| $$  \ $$                                
# |__/      |______/|__/  \__/|______/   |__/    \______/          |__/    \______/  \______/ |________/|_______/  \______/ |__/  |__/                                                                                                                                                                                                                                                                                                                                                                   
                                                                                                                                                                    
#   /$$$$$$  /$$$$$$$  /$$$$$$$$ /$$$$$$$$       /$$$$$$$$ /$$$$$$$$  /$$$$$$  /$$   /$$ /$$   /$$  /$$$$$$  /$$        /$$$$$$   /$$$$$$  /$$$$$$ /$$$$$$$$  /$$$$$$ 
#  /$$__  $$| $$__  $$| $$_____/| $$_____/      |__  $$__/| $$_____/ /$$__  $$| $$  | $$| $$$ | $$ /$$__  $$| $$       /$$__  $$ /$$__  $$|_  $$_/| $$_____/ /$$__  $$
# | $$  \__/| $$  \ $$| $$      | $$               | $$   | $$      | $$  \__/| $$  | $$| $$$$| $$| $$  \ $$| $$      | $$  \ $$| $$  \__/  | $$  | $$      | $$  \__/
# | $$ /$$$$| $$$$$$$/| $$$$$   | $$$$$            | $$   | $$$$$   | $$      | $$$$$$$$| $$ $$ $$| $$  | $$| $$      | $$  | $$| $$ /$$$$  | $$  | $$$$$   |  $$$$$$ 
# | $$|_  $$| $$____/ | $$__/   | $$__/            | $$   | $$__/   | $$      | $$__  $$| $$  $$$$| $$  | $$| $$      | $$  | $$| $$|_  $$  | $$  | $$__/    \____  $$
# | $$  \ $$| $$      | $$      | $$               | $$   | $$      | $$    $$| $$  | $$| $$\  $$$| $$  | $$| $$      | $$  | $$| $$  \ $$  | $$  | $$       /$$  \ $$
# |  $$$$$$/| $$      | $$$$$$$$| $$$$$$$$         | $$   | $$$$$$$$|  $$$$$$/| $$  | $$| $$ \  $$|  $$$$$$/| $$$$$$$$|  $$$$$$/|  $$$$$$/ /$$$$$$| $$$$$$$$|  $$$$$$/
#  \______/ |__/      |________/|________/         |__/   |________/ \______/ |__/  |__/|__/  \__/ \______/ |________/ \______/  \______/ |______/|________/ \______/ 


#  /$$$$$$$$ /$$$$$$ /$$   /$$ /$$$$$$ /$$$$$$$$  /$$$$$$        /$$$$$$$$  /$$$$$$   /$$$$$$  /$$       /$$$$$$$   /$$$$$$  /$$   /$$                                
# | $$_____/|_  $$_/| $$$ | $$|_  $$_/|__  $$__/ /$$__  $$      |__  $$__/ /$$__  $$ /$$__  $$| $$      | $$__  $$ /$$__  $$| $$  / $$                                
# | $$        | $$  | $$$$| $$  | $$     | $$   | $$  \ $$         | $$   | $$  \ $$| $$  \ $$| $$      | $$  \ $$| $$  \ $$|  $$/ $$/                                
# | $$$$$     | $$  | $$ $$ $$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$$$$$$ | $$  | $$ \  $$$$/                                 
# | $$__/     | $$  | $$  $$$$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$__  $$| $$  | $$  >$$  $$                                 
# | $$        | $$  | $$\  $$$  | $$     | $$   | $$  | $$         | $$   | $$  | $$| $$  | $$| $$      | $$  \ $$| $$  | $$ /$$/\  $$                                
# | $$       /$$$$$$| $$ \  $$ /$$$$$$   | $$   |  $$$$$$/         | $$   |  $$$$$$/|  $$$$$$/| $$$$$$$$| $$$$$$$/|  $$$$$$/| $$  \ $$                                
# |__/      |______/|__/  \__/|______/   |__/    \______/          |__/    \______/  \______/ |________/|_______/  \______/ |__/  |__/  

################################################################################
# UNIVERSIDADE FEDERAL DE CATALÃO (UFCAT)
# WANDERLEI MALAQUIAS PEREIRA JUNIOR,                  ENG. CIVIL / PROF (UFCAT)
# MARCOS NAPOLEÃO RABELO,                              ENG. CIVIL / PROF (UFCAT)
# DAVIDSON DE OLIVEIRA FRANÇA JUNIOR,                    ENG. CIVIL / PROF (UNA)
# GABRIEL BERNARDES CARVALHO,                                 ENG. CIVIL (UFCAT)
# JOSÉ VITOR CARVALHO SILVA,                                  ENG. CIVIL (UFCAT)
# MURILO CARNEIRO RODRIGUES,                                  ENG. CIVIL (UFCAT)
################################################################################

################################################################################
# DESCRIÇÃO ALGORITMO:
# BIBLIOTECA FINITO PARA FUNÇÕES COMUNS EM UM ALGORITMO DE ELEMENTOS FINITOS 
# DESENVOLVIDA PELO GRUPO DE PESQUISA E ESTUDOS EM ENGENHARIA (GPEE)
################################################################################

################################################################################
# BIBLIOTECAS NATIVAS PYTHON
import numpy as np

################################################################################
# BIBLIOTECAS DESENVOLVEDORES GPEE

# CRIAÇÃO DOS ÍNDICES INICIAIS PARA EXECUÇÃO DO ALGORITMO
def INDEX_ASSEMBLY(TYPE_ELEMENT):
    """ 
    This function establishes values for the internal variables that are used in the algorithm.
    
    Input:
    TYPE_ELEMENT    | Type element in Finito algorithm           | Integer 
                    |        0 - Frame bar element               |
                    |        1 - CST surface element             |

    Output:
    N_DOFSNODE      | Number of degress of freedom per node      | Integer
    N_NODESELEMENT  | Number of nodes per element                | Integer
    DOFS_LOCAL      | ID local degrees of freedom in element     | Py list[??]
                    |       ?? = [6] -> TYPE_ELEMENT = 0         |
                    |       ?? = [6] -> TYPE_ELEMENT = 1         |
    AUX_1           | Material ID in ELEMENTS                    | Integer
    AUX_2           | Geometry or Thickness ID in ELEMENTS       | Integer
                    |       Geometry  -> TYPE_ELEMENT = 0        |
                    |       Thickness -> TYPE_ELEMENT = 1        |
    N_DOFSELEMENT   | Number of degress of freedom per element   | Integer
    """
    # Frame bar element
    if TYPE_ELEMENT == 0: 
        N_DOFSNODE = 3
        N_NODESELEMENT = 2
        DOFS_LOCAL = [0, 1, 2, 3, 4, 5]
    # CST surface element
    elif TYPE_ELEMENT == 1:
        N_DOFSNODE = 2
        N_NODESELEMENT = 3
        DOFS_LOCAL = [0, 1, 2, 3, 4, 5]
    AUX_1 = N_NODESELEMENT + 0
    AUX_2 = N_NODESELEMENT + 1
    N_DOFSELEMENT = N_DOFSNODE * N_NODESELEMENT
    return N_DOFSNODE, N_NODESELEMENT, DOFS_LOCAL, AUX_1, AUX_2, N_DOFSELEMENT

# DETERMINAÇÃO DOS ID's DOF POR NÓ DA ESTRUTURA
def DOF_GLOBAL_ASSEMBLY(TYPE_ELEMENT, N_DOFSNODE, N_NODES):
    """
    This function determines the value of the degrees of global freedoms by node of the structure.
    
    Input
    TYPE_ELEMENT     | Type element in Finito algorithm           | Integer 
                     |        0 - Frame bar element               |
                     |        1 - CST surface element             |   
    N_DOFSNODE       | Number of degress of freedom per node      | Integer
    N_NODES          | Number of nodes                            | Integer

    Output:
    DOF_GLOBALNODAL  | ID global DOF per node                     | Py Numpy array[N_NODES x ??]
                     |       ?? = [3] -> TYPE_ELEMENT = 0         |
                     |       ?? = [2] -> TYPE_ELEMENT = 1         |
    """
    DOF_GLOBALNODAL = np.zeros((N_NODES, N_DOFSNODE))
    for I_COUNT in range (N_NODES):
        # Frame bar element
        if TYPE_ELEMENT == 0: 
            DOF_GLOBALNODAL[I_COUNT, 0] = int(N_DOFSNODE * I_COUNT + 0)
            DOF_GLOBALNODAL[I_COUNT, 1] = int(N_DOFSNODE * I_COUNT + 1)
            DOF_GLOBALNODAL[I_COUNT, 2] = int(N_DOFSNODE * I_COUNT + 2)
        # CST surface element
        elif TYPE_ELEMENT == 1:
            DOF_GLOBALNODAL[I_COUNT, 0] = int(N_DOFSNODE * I_COUNT + 0)
            DOF_GLOBALNODAL[I_COUNT, 1] = int(N_DOFSNODE * I_COUNT + 1)
    return DOF_GLOBALNODAL

# GRAUS DE LIBERDADE TOTAL
def TOTAL_DEGREE_FREEDOM(N_DOFSNODE, N_NODES):
    """ 
    This function determines the quantity and ID values of the structure's global degrees of freedom.

    Input:
    N_DOFSNODE    | Number of degress of freedom per node  | Integer 
    N_NODES       | Number of nodes                        | Integer 

    Output:
    DOF_GLOBAL    | ID global degree of freedom            | Py list[N_DOFSGLOBAL]
    N_DOFSGLOBAL  | Total of degrees of freedom            | Integer 
    """
    DOF_GLOBAL = []
    N_DOFSGLOBAL = N_NODES * N_DOFSNODE
    for I_COUNT in range (N_DOFSGLOBAL):
        DOF_GLOBAL.append(I_COUNT)
    return DOF_GLOBAL, N_DOFSGLOBAL

# GRAUS DE LIBERDADE RESTRITOS
def PRESCRIPTIONS_DEGREE_FREEDOM(PRESCRIPTIONS, DOF_GLOBALNODAL):
    """
    This function determines the quantity and ID values of the displacements prescribed degrees of freedom. 

    Input: 
    DOF_GLOBALNODAL         | ID global DOF per node                         | Py Numpy array[N_NODES x ??]
                            |       ?? = [3] -> TYPE_ELEMENT = 0             |
                            |       ?? = [2] -> TYPE_ELEMENT = 1             |
    PRESCRIPTIONS           | Prescribed DOF displacement properties         | Py Numpy array              
                            |   Node, Direction (X = 0, Y = 1, Z = 2), Value | 

    Output:
    DOF_PRESCRIPTIONS       | ID prescribed degree of freedom                | Py list[N_DOFSPRESCRIPTIONS]
    DOF_PRESCRIPTIONSVALUE  | Value prescribed degree of freedom             | Py list[N_DOFSPRESCRIPTIONS]
    N_DOFSPRESCRIPTIONS     | Total number of prescribed degrees of freedom  | Integer
    """
    DOF_PRESCRIPTIONS = []
    DOF_PRESCRIPTIONSVALUE = []
    N_DOFSPRESCRIPTIONS = PRESCRIPTIONS.shape[0]
    for I_COUNT in range (N_DOFSPRESCRIPTIONS):
        NODE = int(PRESCRIPTIONS[I_COUNT, 0])
        INDEX_DOF = int(PRESCRIPTIONS[I_COUNT, 1])
        DOF_VALUE = int(DOF_GLOBALNODAL[NODE, INDEX_DOF])
        DOF_PRESCRIPTIONS.append(DOF_VALUE)
        PRESCRIBED_VALUE = PRESCRIPTIONS[I_COUNT, 2]
        DOF_PRESCRIPTIONSVALUE.append(PRESCRIBED_VALUE)
    return DOF_PRESCRIPTIONS, DOF_PRESCRIPTIONSVALUE, N_DOFSPRESCRIPTIONS

# GRAUS DE LIBERDADE LIVRES
def FREE_DEGREE_FREEDOM(DOF_PRESCRIPTIONS, DOF_GLOBAL):
    """
    This function determines the quantity and ID values of the displacements free of degrees of freedom.

    Input:
    DOF_PRESCRIPTIONS  | ID prescribed degree of freedom          | Py list[N_DOFSPRESCRIPTIONS]
    DOF_GLOBAL         | ID global degree of freedom              | Py list[N_DOFSGLOBAL]

    Output:
    DOF_FREE           | ID free degree of freedom                | Py list[N_DOFSFREE] 
    N_DOFSFREE         | Total number of free degrees of freedom  | Integer
    """
    DOF_FREE = np.setdiff1d(DOF_GLOBAL, DOF_PRESCRIPTIONS)
    N_DOFSFREE = len(DOF_FREE)
    return DOF_FREE, N_DOFSFREE

# VETOR COMPLETO DE FORÇAS NODAIS POR DOF
def CONTRIBUTION_NODAL_EXTERNAL_LOAD(NODAL_LOAD, N_DOFSGLOBAL, DOF_GLOBALNODAL):
    """
    This function builds the external charges vector at the level of freedom.

    Input:
    NODAL_LOAD           | Nodal DOF force properties                         | Py Numpy array              
                         |    Node, Direction (X = 0, Y = 1, Z = 2), Value    |
    N_DOFSGLOBAL         | Total of degrees of freedom                        | Integer 
    DOF_GLOBALNODAL      | ID global DOF per node                             | Py Numpy array[N_NODES x ??]
                         |    ?? = [3] -> TYPE_ELEMENT = 0                    |
                         |    ?? = [2] -> TYPE_ELEMENT = 1                    |         
    
    Output:
    DOF_NODALFORCE       | Force per DOF                                      | Py Numpy array[N_DOFSGLOBAL x 1]   
    """
    DOF_NODALFORCE = np.zeros((N_DOFSGLOBAL, 1))
    N_NODALFORCE = NODAL_LOAD.shape[0]
    for I_COUNT in range(N_NODALFORCE):
        NODE = int(NODAL_LOAD[I_COUNT, 0])
        INDEX_DOF = int(NODAL_LOAD[I_COUNT, 1])
        DOF_VALUE = int(DOF_GLOBALNODAL[NODE, INDEX_DOF])
        P = float(NODAL_LOAD[I_COUNT, 2])
        DOF_NODALFORCE[DOF_VALUE, 0] = P
    return DOF_NODALFORCE

# LISTA COM AS CONTRIBUIÇÕES NODAIS DE MOLAS
def SPRING_CONTRIBUTION(N_DOFSNODE, SPRINGS, N_DOFSPRINGS):
    """
    This function creates the contribution of the spring elements in the global stiffness matrix.
    
    Input:
    N_DOFSNODE      | Number of degress of freedom per node           | Integer
    SPRINGS         | Nodal DOF spring properties                     | Py Numpy array              
                    |   Node, Direction (X = 0, Y = 1, Z = 2), Value  |
    N_DOFSPRINGS    | Number of DOF spring elements                   | Integer
    
    Output:
    SPRING_INDEX    | ID DOF spring element                           | Py list[N_DOFSPRINGS]
    SPRING_VALUES   | Spring coefficient                              | Integer 
    """
    SPRING_INDEX = []
    SPRING_VALUES = []
    for I_COUNT in range(N_DOFSPRINGS):
        NODE = int(SPRINGS[I_COUNT, 0])
        INDEX_DOF = int(SPRINGS[I_COUNT, 1])
        SPRING_INDEX.append(int(N_DOFSNODE * NODE + INDEX_DOF))
        SPRING_VALUES.append(SPRINGS[I_COUNT, 2])
    return SPRING_INDEX, SPRING_VALUES

# DOF GLOBAL POR ELEMENTO
def GLOBAL_DOF_ELEMENT(N_NODESELEMENT, N_DOFSNODE, DOF_GLOBALNODAL, ELEMENTS, I_ELEMENT):
    """
    This function determines the degree of global freedom of the element I_ELEMENT.

    Input:
    N_NODESELEMENT      | Number of nodes per element                            | Integer
    N_DOFSNODE          | Number of degress of freedom per node                  | Integer
    DOF_GLOBALNODAL     | ID global DOF per node                                 | Py Numpy array[N_NODES x ??]
                        |    ?? = [3] -> TYPE_ELEMENT = 0                        |
                        |    ?? = [2] -> TYPE_ELEMENT = 1                        |   
    ELEMENTS            | Elements properties                                    | Py Numpy array
                        |   Node 0 ... Node (N_NODES - 1), Material ID,          |
                        |    Geometry ID, Hinge ID node 0, Hinge ID node 1       |
    I_ELEMENT           | i element in looping                                   | Integer
    
    Output:
    DOF_GLOBALIELEMENT  | Global DOF ID I_ELEMENT                                | Py list[N_DOFSELEMENT] 
    """
    DOF_GLOBALIELEMENT = []
    for I_COUNT in range(N_NODESELEMENT):
        NODE = int(ELEMENTS[I_ELEMENT, I_COUNT])
        for J_COUNT in range(N_DOFSNODE):
            DOF_VALUE = int(DOF_GLOBALNODAL[NODE, J_COUNT])
            DOF_GLOBALIELEMENT.append(DOF_VALUE)
    return DOF_GLOBALIELEMENT

# MATRIZ DE RIGIDEZ GLOBAL
def GLOBAL_STIFFNESS(N_DOFSGLOBAL, DOF_GLOBALIELEMENT, K_IELEMENT):
    """
    This function creates the global stiffness.

    Input:
    N_DOFSGLOBAL        | Total of degrees of freedom       | Integer 
    DOF_GLOBALIELEMENT  | Global DOF ID I_ELEMENT           | Integer 
    K_IELEMENT          | Local stiffness matrix I_ELEMENT  | Py Numpy array[N_DOFSELEMENT x N_DOFSELEMENT]

    Output:
    K_G                 | Global stiffness matrix           | Py Numpy array[N_DOFSGLOBAL x N_DOFSGLOBAL]
    """
    K_G = np.zeros((N_DOFSGLOBAL, N_DOFSGLOBAL))
    for I_COUNT, I_VALUE in enumerate(DOF_GLOBALIELEMENT):
        for J_COUNT, J_VALUE in enumerate(DOF_GLOBALIELEMENT):
            K_G[I_VALUE, J_VALUE] = K_G[I_VALUE, J_VALUE] + K_IELEMENT[I_COUNT, J_COUNT]
    return K_G

# MATRIZ DE RIGIDEZ COM OS DOF's LIVRES 
def CONDENSE_GLOBAL_FREE_STIFFNESS(K_G, DOF_FREE, N_DOFSFREE):
    """ 
    This function assembles the portion containing the free degrees of freedom of the stiffness matrix. 
    
    Input:
    K_G         | Global stiffness matrix                  | Py Numpy array[N_DOFSGLOBAL x N_DOFSGLOBAL]
    DOF_FREE    | ID free degree of freedom                | Py list[N_DOFSFREE]     
    N_DOFSFREE  | Total number of free degrees of freedom  | Integer
    
    Output:
    K_FF        | Free global stiffness matrix             | Py Numpy array[N_DOFSFREE x N_DOFSFREE]
    """
    K_FF = np.zeros((N_DOFSFREE, N_DOFSFREE))
    for I_COUNT in range(N_DOFSFREE):
        DOF_FREELINE = DOF_FREE[I_COUNT]
        for J_COUNT in range(N_DOFSFREE):
            DOF_FREECOLUMN = DOF_FREE[J_COUNT]
            K_FF[I_COUNT, J_COUNT] = K_G[DOF_FREELINE, DOF_FREECOLUMN]
    return K_FF

# MATRIZ DE RIGIDEZ COM OS DOF's PRESCRITOS-LIVRES  
def CONDENSE_PRESCRIBED_FREE_GLOBAL_STIFFNESS(K_G, DOF_FREE, N_DOFSFREE, DOF_PRESCRIPTIONS, N_DOFSPRESCRIPTIONS):
    """ 
    This function assembles the portion containing the prescribed and free degrees of freedom of the stiffness matrix.
    
    Input:
    K_G                  | Global stiffness matrix                        | Py Numpy array[N_DOFSGLOBAL x N_DOFSGLOBAL]
    DOF_FREE             | ID free degree of freedom                      | Py list[N_DOFSFREE]     
    N_DOFSFREE           | Total number of free degrees of freedom        | Integer
    DOF_PRESCRIPTIONS    | ID prescribed degree of freedom                | Py list[N_DOFSPRESCRIPTIONS]
    N_DOFSPRESCRIPTIONS  | Total number of prescribed degrees of freedom  | Integer
    
    Output:
    K_PF                 | Prescribed-Free global stiffness matrix        | Py Numpy array[N_DOFSPRESCRIPTIONS x N_DOFSFREE]
    """
    K_PF = np.zeros((N_DOFSPRESCRIPTIONS, N_DOFSFREE))
    for I_COUNT in range(N_DOFSPRESCRIPTIONS):
        DOF_PRECRIBEDLINE = DOF_PRESCRIPTIONS[I_COUNT]
        for J_COUNT in range(N_DOFSFREE):
            DOF_FREECOLUMN = DOF_FREE[J_COUNT]
            K_PF[I_COUNT, J_COUNT] = K_G[DOF_PRECRIBEDLINE, DOF_FREECOLUMN]
    return K_PF

# VETOR GLOBAL DE DESLOCAMENTOS COM OS DOF's PRESCRITOS 
def CONDENSE_PRESCRIBED_GLOBAL_DISPLACEMENT(DOF_PRESCRIPTIONSVALUE, N_DOFSPRESCRIPTIONS):
    """ 
    This function assembles the portion containing the prescribed degrees of freedom of the global displacement.
    
    Input:
    DOF_PRESCRIPTIONSVALUE  | Value prescribed degree of freedom             | Py list[N_DOFSPRESCRIPTIONS]
    N_DOFSPRESCRIPTIONS     | Total number of prescribed degrees of freedom  | Integer
    
    Output:
    U_PP                    | Prescribed global displacement                 | Py Numpy array[N_DOFSPRESCRIPTIONS x 1]
    """
    U_PP = np.zeros((N_DOFSPRESCRIPTIONS, 1))
    for I_COUNT in range(N_DOFSPRESCRIPTIONS):
        DOF_PRESCRIBEDVALUE = DOF_PRESCRIPTIONSVALUE[I_COUNT]
        U_PP[I_COUNT, 0] = DOF_PRESCRIBEDVALUE
    return U_PP

# VETOR GLOBAL DE FORÇAS COM OS DOF's LIVRES
def CONDENSE_FREE_GLOBAL_FORCES(F_G, DOF_FREE, N_DOFSFREE):
    """ 
    This function assembles the portion containing the free degrees 
    of freedom of the global forces.
    
    Input:
    F_G         | Global forces                             | Py Numpy array[N_DOFSGLOBAL x 1]
    DOF_FREE    | ID free degree of freedom                 | Py list[N_DOFSFREE]    
    N_DOFSFREE  | Total number of free degrees of freedom   | Integer
    
    Output:
    F_FF        | Free global forces                        | Py Numpy array[N_DOFSPRESCRIPTIONS x 1]
    """    
    F_FF = np.zeros((N_DOFSFREE, 1))
    for I_COUNT in range(N_DOFSFREE):
        FREE_DOF = DOF_FREE[I_COUNT]
        F_FF[I_COUNT, 0] = F_G[FREE_DOF, 0]
    return F_FF

# MONTAGEM DO VETOR DE DESLOCAMENTOS GLOBAIS DA ESTRUTURA
def ASSEMBLY_TOTAL_DISPLACEMENT(U_FF, U_PP, N_DOFSGLOBAL, DOF_PRESCRIPTIONS, DOF_FREE):
    """
    This function assembles the global displacements from the calculated parcels.
    
    Input:
    U_FF                 | Free global displacement         | Py Numpy array[N_DOFSFREE x 1]
    U_PP                 | Prescribed global displacement   | Py Numpy array[N_DOFSPRESCRIPTIONS x 1]
    N_DOFSGLOBAL         | Total of degrees of freedom      | Integer 
    DOF_PRESCRIPTIONS    | ID prescribed degree of freedom  | Py list[N_DOFSPRESCRIPTIONS]    
    DOF_FREE             | ID free degree of freedom        | Py list[N_DOFSFREE]     
    
    Output:
    U_G                  | Global displacement              | Py Numpy array[N_DOFSFREE x 1]
    """
    U_G = np.zeros((N_DOFSGLOBAL, 1))
    for I_COUNT, I_VALUE in enumerate(DOF_PRESCRIPTIONS):
        DOF_DISPLACEMENTVALUE = U_PP[I_COUNT, 0]
        U_G[I_VALUE, 0] = DOF_DISPLACEMENTVALUE
    for J_COUNT, J_VALUE in enumerate(DOF_FREE):
        DOF_DISPLACEMENTVALUE = U_FF[J_COUNT, 0]
        U_G[J_VALUE, 0] = DOF_DISPLACEMENTVALUE
    return U_G

# DESLOCAMENTOS GLOBAIS POR ELEMENTO
def CONDENSE_GLOBAL_ELEMENT_DISPLACEMENTS(U_G, N_DOFSELEMENT, DOF_GLOBALIELEMENT):
    """
    This function assembles the nodal displacements of I_ELEMENT.

    Input:
    U_G                 | Global displacement                        | Py Numpy array[N_DOFSFREE x 1]
    N_DOFSELEMENT       | Number of degress of freedom per element   | Integer
    DOF_GLOBALIELEMENT  | Global DOF ID I_ELEMENT                    | Py list[N_DOFSELEMENT] 
    
    Output:    
    U_IELEMENT          | Global displacements I_ELEMENT             | Py Numpy array[N_DOFSELEMENT x 1] 
    """
    U_IELEMENT = np.zeros((N_DOFSELEMENT, 1))
    for I_COUNT, J_COUNT in enumerate(DOF_GLOBALIELEMENT):
        U_IELEMENT[I_COUNT, 0] = U_G[J_COUNT, 0]
    return U_IELEMENT

# FORÇAS INTERNAS DA ESTRUTURA
def GLOBAL_INTERNAL_LOADS(F_INTIELEMENT, N_DOFSGLOBAL, DOF_GLOBALIELEMENT):
    """
    This function assembles the internal load of the structure.

    Input:
    F_INTIELEMENT       | Global internal load I_ELEMENT  | Py Numpy array[N_DOFSELEMENT x 1] 
    N_DOFSGLOBAL        | Total of degrees of freedom     | Integer 
    DOF_GLOBALIELEMENT  | Global DOF ID I_ELEMENT         | Py list[N_DOFSELEMENT] 
    
    Output:
    F_INT               | Global internal load            | Py Numpy array[N_DOFSGLOBA x 1] 
    """
    F_INT = np.zeros((N_DOFSGLOBAL, 1))
    for I_COUNT, J_COUNT in enumerate(DOF_GLOBALIELEMENT):
        F_INT[J_COUNT, 0] = F_INTIELEMENT[I_COUNT, 0]
    return F_INT

# IMPOSIÇÃO DAS CONDIÇÕES DE CONTORNO TÉCNICA 0 - 1
def ZERO_AND_ONE_METHOD(K_G, F_G, DOF_PRESCRIPTIONS, DOF_PRESCRIPTIONSVALUE):
    """
    This function solves the system of linear equations using the 0 and 1 technique.
    
    Input:
    K_G                     | Global stiffness matrix             | Py Numpy array[N_DOFSGLOBAL x N_DOFSGLOBAL]
    F_G                     | Global forces                       | Py Numpy array[N_DOFSGLOBAL x 1]
    DOF_PRESCRIPTIONS       | ID prescribed degree of freedom     | Py list[N_DOFSPRESCRIPTIONS]  
    DOF_PRESCRIPTIONSVALUE  | Value prescribed degree of freedom  | Py list[N_DOFSPRESCRIPTIONS]

    Output:
    K_G                     | Update global stiffness matrix      | Py Numpy array[N_DOFSGLOBAL x N_DOFSGLOBAL]
    F_G                     | Update global forces                | Py Numpy array[N_DOFSGLOBAL x 1]
    """
    for I_COUNT, DOF in enumerate(DOF_PRESCRIPTIONS):
        VALUE_PRESCRIBED = DOF_PRESCRIPTIONSVALUE[I_COUNT]
        AUX_1 = K_G[:, DOF]
        AUX_1 = np.expand_dims(AUX_1, axis = 1)
        F_G -= VALUE_PRESCRIBED * AUX_1
        F_G[DOF, 0] = VALUE_PRESCRIBED
        K_G[DOF,:] = 0
        K_G[:, DOF] = 0
        K_G[DOF, DOF] = 1
    return K_G, F_G                                                                                                                                                                                                                                                                                                                                                                
                                                                                                                                                                    
#   /$$$$$$  /$$$$$$$  /$$$$$$$$ /$$$$$$$$       /$$$$$$$$ /$$$$$$$$  /$$$$$$  /$$   /$$ /$$   /$$  /$$$$$$  /$$        /$$$$$$   /$$$$$$  /$$$$$$ /$$$$$$$$  /$$$$$$ 
#  /$$__  $$| $$__  $$| $$_____/| $$_____/      |__  $$__/| $$_____/ /$$__  $$| $$  | $$| $$$ | $$ /$$__  $$| $$       /$$__  $$ /$$__  $$|_  $$_/| $$_____/ /$$__  $$
# | $$  \__/| $$  \ $$| $$      | $$               | $$   | $$      | $$  \__/| $$  | $$| $$$$| $$| $$  \ $$| $$      | $$  \ $$| $$  \__/  | $$  | $$      | $$  \__/
# | $$ /$$$$| $$$$$$$/| $$$$$   | $$$$$            | $$   | $$$$$   | $$      | $$$$$$$$| $$ $$ $$| $$  | $$| $$      | $$  | $$| $$ /$$$$  | $$  | $$$$$   |  $$$$$$ 
# | $$|_  $$| $$____/ | $$__/   | $$__/            | $$   | $$__/   | $$      | $$__  $$| $$  $$$$| $$  | $$| $$      | $$  | $$| $$|_  $$  | $$  | $$__/    \____  $$
# | $$  \ $$| $$      | $$      | $$               | $$   | $$      | $$    $$| $$  | $$| $$\  $$$| $$  | $$| $$      | $$  | $$| $$  \ $$  | $$  | $$       /$$  \ $$
# |  $$$$$$/| $$      | $$$$$$$$| $$$$$$$$         | $$   | $$$$$$$$|  $$$$$$/| $$  | $$| $$ \  $$|  $$$$$$/| $$$$$$$$|  $$$$$$/|  $$$$$$/ /$$$$$$| $$$$$$$$|  $$$$$$/
#  \______/ |__/      |________/|________/         |__/   |________/ \______/ |__/  |__/|__/  \__/ \______/ |________/ \______/  \______/ |______/|________/ \______/ 


from setuptools import setup

setup(
    name = 'FINITO_TOOLBOX',
    version = '2022.1',
    author = 'Prof. Wanderlei M. Pereira Junior, Prof. Davidson de Oliveira França Junior, Prof. Marcos Rabelo Napoleão, Civil Eng. Gabriel B. Carvalho, Civil Eng. José Vitor Carvalho and Civil Eng. Murilo Carneiro Rodrigues',
    author_email = 'wanderlei_junior@ufcat.edu.br',
    url='https://wmpjrufg.github.io/FINITO_TOOLBOX/',
    packages = ['FINITO_TOOLBOX']
)